import ipaddress
from ipaddress import IPv4Interface
from typing import TYPE_CHECKING, Any

from graphene import InputObjectType, Mutation
from graphql import GraphQLResolveInfo
from typing_extensions import Self

from infrahub import lock
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import InfrahubKind, MetadataOptions
from infrahub.core.file_processor import FileUploadProcessor
from infrahub.core.ipam.reconciler import IpamReconciler
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.lock import InfrahubMultiLock
from infrahub.log import get_logger

from ...core.node.create import create_node
from ...core.node.lock_utils import build_object_lock_name, get_lock_names_on_object_mutation
from .main import DeleteResult, InfrahubMutationMixin, InfrahubMutationOptions, UpsertResult, build_graphql_response
from .node_getter.by_default_filter import MutationNodeGetterByDefaultFilter

if TYPE_CHECKING:
    from infrahub.graphql.initialization import GraphqlContext

log = get_logger()


async def validate_namespace(
    db: InfrahubDatabase,
    branch: Branch | str | None,
    data: InputObjectType,
    existing_namespace_id: str | None = None,
) -> str:
    """Validate or set (if not present) the namespace to pass to the mutation and return its ID."""
    namespace_id: str | None = None
    if "ip_namespace" not in data or not data["ip_namespace"]:
        namespace_id = existing_namespace_id or registry.default_ipnamespace
        data["ip_namespace"] = {"id": namespace_id}
    elif "id" in data["ip_namespace"]:
        namespace = await registry.manager.get_one(
            db=db, branch=branch, kind=InfrahubKind.IPNAMESPACE, id=data["ip_namespace"]["id"]
        )
        namespace_id = namespace.id
    elif "hfid" in data["ip_namespace"]:
        namespace = await registry.manager.get_one_by_hfid(
            db=db, branch=branch, kind=InfrahubKind.IPNAMESPACE, hfid=data["ip_namespace"]["hfid"], raise_on_error=True
        )
        namespace_id = namespace.id
    else:
        raise ValidationError(
            "A valid ip_namespace must be provided or ip_namespace should be left empty in order to use the default value."
        )
    return namespace_id


class InfrahubIPNamespaceMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema,
        _meta: Any | None = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> DeleteResult:
        if data["id"] == registry.default_ipnamespace:
            raise ValueError("Cannot delete default IPAM namespace")

        return await super().mutate_delete(info=info, data=data, branch=branch)


class InfrahubIPAddressMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema,
        _meta: Any | None = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @staticmethod
    def _get_lock_names(namespace_id: str, branch: Branch) -> list[str]:
        if not branch.is_default:
            # Do not lock on other branches as reconciliation will be performed at least when merging in main branch.
            return []
        return [build_object_lock_name(InfrahubKind.IPADDRESS + "_" + namespace_id)]

    @classmethod
    async def _mutate_create_object_and_reconcile(
        cls,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        ip_address: IPv4Interface | ipaddress.IPv6Interface,
        namespace_id: str,
    ) -> Node:
        address = await create_node(
            data=dict(data),
            db=db,
            branch=branch,
            schema=cls._meta.active_schema,
        )

        reconciler = IpamReconciler(db=db, branch=branch)
        reconciled_address = await reconciler.reconcile(
            ip_value=ip_address, namespace=namespace_id, node_uuid=address.get_id()
        )
        return reconciled_address

    @classmethod
    @retry_db_transaction(name="ipaddress_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        ip_address = ipaddress.ip_interface(data["address"]["value"])
        namespace_id = await validate_namespace(db=db, branch=branch, data=data)

        lock_names = cls._get_lock_names(namespace_id, branch)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
            async with db.start_transaction() as dbt:
                reconciled_address = await cls._mutate_create_object_and_reconcile(
                    data=data, branch=branch, db=dbt, ip_address=ip_address, namespace_id=namespace_id
                )
                graphql_response = await build_graphql_response(info=info, db=dbt, obj=reconciled_address)

        return reconciled_address, cls(**graphql_response)

    @classmethod
    async def _mutate_update_object_and_reconcile(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        address: Node,
        namespace_id: str,
    ) -> Node:
        address = await cls.mutate_update_object(db=db, info=info, data=data, branch=branch, obj=address)
        reconciler = IpamReconciler(db=db, branch=branch)
        ip_address = ipaddress.ip_interface(address.address.value)
        reconciled_address = await reconciler.reconcile(
            ip_value=ip_address, node_uuid=address.get_id(), namespace=namespace_id
        )
        return reconciled_address

    @classmethod
    @retry_db_transaction(name="ipaddress_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        address = node or await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=cls._meta.schema.kind,
            id=data.get("id"),
            branch=branch,
            include_metadata=MetadataOptions.LINKED_NODES,
        )
        namespace = await address.ip_namespace.get_peer(db)
        namespace_id = await validate_namespace(db=db, branch=branch, data=data, existing_namespace_id=namespace.id)

        # Prepare a clone to compute locks without triggering pool allocations
        preview_obj = await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=address.get_kind(),
            id=address.get_id(),
            branch=branch,
        )
        await preview_obj.from_graphql(db=db, data=data, process_pools=False)

        schema_branch = db.schema.get_schema_branch(name=branch.name)
        lock_names = get_lock_names_on_object_mutation(node=preview_obj, schema_branch=schema_branch)

        namespace_lock_names = cls._get_lock_names(namespace_id, branch)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=namespace_lock_names):
            # FIXME: do not lock when data does not contain uniqueness constraint fields or resource pool allocations
            async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names, metrics=False):
                async with db.start_transaction() as dbt:
                    reconciled_address = await cls._mutate_update_object_and_reconcile(
                        info=info, data=data, branch=branch, address=address, namespace_id=namespace_id, db=dbt
                    )

                    result = await cls.mutate_update_to_graphql(db=dbt, info=info, obj=reconciled_address)

        return address, result

    @classmethod
    async def mutate_upsert(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        node_getter_default_filter: MutationNodeGetterByDefaultFilter,
        database: InfrahubDatabase | None = None,
        file_processor: FileUploadProcessor | None = None,
    ) -> UpsertResult:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        await validate_namespace(db=db, branch=branch, data=data)
        return await super().mutate_upsert(
            info=info,
            data=data,
            branch=branch,
            node_getter_default_filter=node_getter_default_filter,
            database=db,
            file_processor=file_processor,
        )

    @classmethod
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> DeleteResult:
        return await super().mutate_delete(info=info, data=data, branch=branch)


class InfrahubIPPrefixMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema,
        _meta: Any | None = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @staticmethod
    def _get_lock_names(namespace_id: str) -> list[str]:
        # IPPrefix has some cardinality-one relationships involved (parent/child/ip_address),
        # so we need to lock on any branch to avoid creating multiple peers for these relationships
        # during concurrent ipam reconciliations.
        return [build_object_lock_name(InfrahubKind.IPPREFIX + "_" + namespace_id)]

    @classmethod
    async def _mutate_create_object_and_reconcile(
        cls,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        namespace_id: str,
    ) -> Node:
        prefix = await create_node(
            data=dict(data),
            db=db,
            branch=branch,
            schema=cls._meta.active_schema,
        )
        return await cls._reconcile_prefix(
            branch=branch, db=db, prefix=prefix, namespace_id=namespace_id, is_delete=False
        )

    @classmethod
    @retry_db_transaction(name="ipprefix_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        namespace_id = await validate_namespace(db=db, branch=branch, data=data)

        lock_names = cls._get_lock_names(namespace_id)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
            async with db.start_transaction() as dbt:
                reconciled_prefix = await cls._mutate_create_object_and_reconcile(
                    data=data, branch=branch, db=dbt, namespace_id=namespace_id
                )

            graphql_response = await build_graphql_response(info=info, db=dbt, obj=reconciled_prefix)

        return reconciled_prefix, cls(**graphql_response)

    @classmethod
    async def _mutate_update_object_and_reconcile(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        prefix: Node,
        namespace_id: str,
    ) -> Node:
        prefix = await cls.mutate_update_object(db=db, info=info, data=data, branch=branch, obj=prefix)
        return await cls._reconcile_prefix(
            branch=branch, db=db, prefix=prefix, namespace_id=namespace_id, is_delete=False
        )

    @classmethod
    @retry_db_transaction(name="ipprefix_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        prefix = node or await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=cls._meta.schema.kind,
            id=data.get("id"),
            branch=branch,
            include_metadata=MetadataOptions.LINKED_NODES,
        )
        namespace = await prefix.ip_namespace.get_peer(db)
        namespace_id = await validate_namespace(db=db, branch=branch, data=data, existing_namespace_id=namespace.id)

        # Prepare a clone to compute locks without triggering pool allocations
        preview_obj = await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=prefix.get_kind(),
            id=prefix.get_id(),
            branch=branch,
        )
        await preview_obj.from_graphql(db=db, data=data, process_pools=False)

        schema_branch = db.schema.get_schema_branch(name=branch.name)
        lock_names = get_lock_names_on_object_mutation(node=preview_obj, schema_branch=schema_branch)

        namespace_lock_names = cls._get_lock_names(namespace_id)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=namespace_lock_names):
            async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names, metrics=False):
                async with db.start_transaction() as dbt:
                    reconciled_prefix = await cls._mutate_update_object_and_reconcile(
                        info=info, data=data, prefix=prefix, db=dbt, namespace_id=namespace_id, branch=branch
                    )
                    result = await cls.mutate_update_to_graphql(db=dbt, info=info, obj=reconciled_prefix)

        return prefix, result

    @classmethod
    async def mutate_upsert(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        node_getter_default_filter: MutationNodeGetterByDefaultFilter,
        database: InfrahubDatabase | None = None,
        file_processor: FileUploadProcessor | None = None,
    ) -> UpsertResult:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        await validate_namespace(db=db, branch=branch, data=data)
        return await super().mutate_upsert(
            info=info,
            data=data,
            branch=branch,
            node_getter_default_filter=node_getter_default_filter,
            database=db,
            file_processor=file_processor,
        )

    @classmethod
    async def _reconcile_prefix(
        cls,
        branch: Branch,
        db: InfrahubDatabase,
        prefix: Node,
        namespace_id: str,
        is_delete: bool,
    ) -> Node:
        reconciler = IpamReconciler(db=db, branch=branch)
        ip_network = ipaddress.ip_network(prefix.prefix.value)
        reconciled_prefix = await reconciler.reconcile(
            ip_value=ip_network, node_uuid=prefix.get_id(), namespace=namespace_id, is_delete=is_delete
        )
        return reconciled_prefix

    @classmethod
    @retry_db_transaction(name="ipprefix_delete")
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> DeleteResult:
        graphql_context: GraphqlContext = info.context
        db = graphql_context.db

        prefix = await NodeManager.get_one(
            data.get("id"), graphql_context.db, branch=branch, prefetch_relationships=True
        )
        if not prefix:
            raise NodeNotFoundError(node_type=cls._meta.schema.kind, identifier=data.get("id"), branch_name=branch.name)

        namespace_rels = await prefix.ip_namespace.get_relationships(db=db)
        namespace_id = namespace_rels[0].peer_id

        lock_names = cls._get_lock_names(namespace_id)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
            async with graphql_context.db.start_transaction() as dbt:
                reconciled_prefix = await cls._reconcile_prefix(
                    branch=branch, db=dbt, prefix=prefix, namespace_id=namespace_id, is_delete=True
                )
        ok = True

        return DeleteResult(node=reconciled_prefix, mutation=cls(ok=ok))



from __future__ import annotations

from typing import TYPE_CHECKING

from prefect import flow
from prefect.logging import get_run_logger

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.trigger.models import TriggerSetupReport, TriggerType
from infrahub.trigger.setup import setup_triggers_specific
from infrahub.workers.dependencies import get_client, get_component, get_database, get_workflow
from infrahub.workflows.catalogue import PROFILE_REFRESH
from infrahub.workflows.utils import add_tags, wait_for_schema_to_converge

from .gather import gather_trigger_profile_refresh

if TYPE_CHECKING:
    from infrahub_sdk.node.relationship import RelationshipManager

REFRESH_PROFILES_MUTATION = """
mutation RefreshProfiles(
    $id: String!,
  ) {
  InfrahubProfilesRefresh(
    data: {id: $id}
  ) {
    ok
  }
}
"""


@flow(name="object-profiles-refresh", flow_run_name="Refresh profiles for {node_id}")
async def object_profiles_refresh(branch_name: str, node_id: str) -> None:
    log = get_run_logger()
    client = get_client()

    await add_tags(branches=[branch_name], nodes=[node_id], db_change=True)
    await client.execute_graphql(query=REFRESH_PROFILES_MUTATION, variables={"id": node_id}, branch_name=branch_name)
    log.info(f"Profiles refreshed for {node_id}")


@flow(name="objects-profiles-refresh-multiple", flow_run_name="Refresh profiles for multiple objects")
async def objects_profiles_refresh_multiple(branch_name: str, node_ids: list[str]) -> None:
    log = get_run_logger()

    await add_tags(branches=[branch_name])

    for node_id in node_ids:
        log.info(f"Requesting profile refresh for {node_id}")
        await get_workflow().submit_workflow(
            workflow=PROFILE_REFRESH, parameters={"branch_name": branch_name, "node_id": node_id}
        )


@flow(name="profile-refresh-setup", flow_run_name="Setup profile refresh triggers")
async def profile_refresh_setup(
    context: InfrahubContext,  # noqa: ARG001
    branch_name: str | None = None,
    event_name: str | None = None,  # noqa: ARG001
) -> None:
    """Setup Prefect automations for profile refresh triggers.

    This flow is triggered by schema changes and sets up automations that will
    listen for profile updates. When a profile's attributes or relationships
    change, the corresponding automation will trigger profile refresh for all
    related nodes.
    """
    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()

        if branch_name:
            await add_tags(branches=[branch_name])
            component = await get_component()
            await wait_for_schema_to_converge(branch_name=branch_name, component=component, db=db, log=log)

        report: TriggerSetupReport = await setup_triggers_specific(
            gatherer=gather_trigger_profile_refresh, trigger_type=TriggerType.PROFILE
        )  # type: ignore[misc]

        log.info(f"{report.in_use_count} Profile refresh automation configuration completed")


@flow(name="profile-refresh-process", flow_run_name="Process profile refresh for {profile_kind}")
async def profile_refresh_process(
    branch_name: str,
    profile_kind: str,
    profile_id: str,
    context: InfrahubContext,  # noqa: ARG001
) -> None:
    """Process profile refresh when a profile's attributes or relationships change.

    This flow fetches all nodes related to the profile via the `related_nodes`
    relationship and submits profile refresh workflows for each of them.
    """
    log = get_run_logger()
    client = get_client()

    await add_tags(branches=[branch_name])

    profile = await client.get(kind=profile_kind, id=profile_id, branch=branch_name, include=["related_nodes"])
    related_nodes: RelationshipManager = profile.related_nodes  # type: ignore

    if not related_nodes.peer_ids:
        log.info(f"No related nodes found for profile {profile_id}")
        return

    log.info(f"Found {len(related_nodes.peer_ids)} related nodes for profile {profile_id}")

    for node_id in related_nodes.peer_ids:
        log.info(f"Requesting profile refresh for {node_id}")
        await get_workflow().submit_workflow(
            workflow=PROFILE_REFRESH, parameters={"branch_name": branch_name, "node_id": node_id}
        )



import importlib
from collections.abc import Sequence
from typing import TYPE_CHECKING
from uuid import uuid4

from infrahub import config, lock
from infrahub.constants.database import DatabaseType
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.branch.enums import BranchStatus
from infrahub.core.constants import (
    DEFAULT_IP_NAMESPACE,
    GLOBAL_BRANCH_NAME,
    GlobalPermissions,
    InfrahubKind,
    PermissionAction,
    PermissionDecision,
)
from infrahub.core.graph import GRAPH_VERSION
from infrahub.core.graph.index import node_indexes, rel_indexes
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.node.ipam import BuiltinIPPrefix
from infrahub.core.node.permissions import CoreGlobalPermission, CoreObjectPermission
from infrahub.core.node.proposed_change import CoreProposedChange
from infrahub.core.node.resource_manager.ip_address_pool import CoreIPAddressPool
from infrahub.core.node.resource_manager.ip_prefix_pool import CoreIPPrefixPool
from infrahub.core.node.resource_manager.number_pool import CoreNumberPool
from infrahub.core.protocols import CoreAccount, CoreAccountGroup, CoreAccountRole
from infrahub.core.root import Root
from infrahub.core.schema import SchemaRoot, core_models, internal_schema
from infrahub.core.schema.manager import SchemaManager
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase
from infrahub.database.memgraph import IndexManagerMemgraph
from infrahub.database.neo4j import IndexManagerNeo4j
from infrahub.exceptions import DatabaseError
from infrahub.graphql.manager import registry as graphql_registry
from infrahub.log import get_logger
from infrahub.menu.utils import create_default_menu
from infrahub.permissions import PermissionBackend, get_or_create_global_permission
from infrahub.storage import InfrahubObjectStorage

if TYPE_CHECKING:
    from infrahub.database.index import IndexManagerBase

log = get_logger()


async def get_root_node(db: InfrahubDatabase, initialize: bool = False) -> Root:
    roots = await Root.get_list(db=db)
    if len(roots) == 0 and not initialize:
        raise DatabaseError(
            "The Database hasn't been initialized for Infrahub, please 'infrahub server start' to initialize the database."
        )

    if len(roots) == 0:
        await first_time_initialization(db=db)
        roots = await Root.get_list(db=db)

    elif len(roots) > 1:
        raise DatabaseError("The Database is corrupted, more than 1 root node found.")

    return roots[0]


async def get_default_ipnamespace(db: InfrahubDatabase) -> Node | None:
    if not registry.schema._branches or not registry.schema.has(name=InfrahubKind.NAMESPACE):
        return None

    nodes = await registry.manager.query(db=db, schema=InfrahubKind.NAMESPACE, filters={"default__value": True})
    if len(nodes) == 0:
        return None

    if len(nodes) > 1:
        raise DatabaseError("More than 1 default namespace found.")

    return nodes[0]


def initialize_permission_backends() -> list[PermissionBackend]:
    permission_backends: list[PermissionBackend] = []
    for backend_module_path in config.SETTINGS.main.permission_backends:
        log.info("Loading permission backend", backend=backend_module_path)

        module, class_name = backend_module_path.rsplit(".", maxsplit=1)
        Backend = getattr(importlib.import_module(module), class_name)
        permission_backends.append(Backend())

    return permission_backends


async def initialize_registry(db: InfrahubDatabase, initialize: bool = False) -> None:
    # ---------------------------------------------------
    # Initialize the database and Load the Root node
    # ---------------------------------------------------
    root = await get_root_node(db=db, initialize=initialize)
    registry.id = str(root.get_uuid())
    registry.default_branch = root.default_branch

    # ---------------------------------------------------
    # Initialize the Storage Driver
    # ---------------------------------------------------
    registry.storage = await InfrahubObjectStorage.init(settings=config.SETTINGS.storage)

    # ---------------------------------------------------
    # Load existing branches into the registry
    # ---------------------------------------------------
    branches: list[Branch] = await Branch.get_list(db=db)
    for branch in branches:
        registry.branch[branch.name] = branch

    # ---------------------------------------------------
    # Load internal models into the registry
    # ---------------------------------------------------
    registry.node["Node"] = Node
    registry.node[InfrahubKind.IPPREFIX] = BuiltinIPPrefix
    registry.node[InfrahubKind.IPADDRESSPOOL] = CoreIPAddressPool
    registry.node[InfrahubKind.IPPREFIXPOOL] = CoreIPPrefixPool
    registry.node[InfrahubKind.NUMBERPOOL] = CoreNumberPool
    registry.node[InfrahubKind.GLOBALPERMISSION] = CoreGlobalPermission
    registry.node[InfrahubKind.OBJECTPERMISSION] = CoreObjectPermission
    registry.node[InfrahubKind.PROPOSEDCHANGE] = CoreProposedChange

    # ---------------------------------------------------
    # Instantiate permission backends
    # ---------------------------------------------------
    registry.permission_backends = initialize_permission_backends()


async def add_indexes(db: InfrahubDatabase) -> None:
    if db.db_type is DatabaseType.MEMGRAPH:
        index_manager: IndexManagerBase = IndexManagerMemgraph(db=db)
    index_manager = IndexManagerNeo4j(db=db)

    index_manager.init(nodes=node_indexes, rels=rel_indexes)
    log.debug("Loading database indexes ..")
    await index_manager.add()


async def initialization(db: InfrahubDatabase, add_database_indexes: bool = False) -> bool:
    """Run initialization and setup, returns a boolean to indicate if it's the initial setup."""
    if config.SETTINGS.database.db_type == config.DatabaseType.MEMGRAPH:
        session = await db.session()
        await session.run(query="SET DATABASE SETTING 'log.level' TO 'INFO'")
        await session.run(query="SET DATABASE SETTING 'log.to_stderr' TO 'true'")
        await session.run(query="STORAGE MODE IN_MEMORY_ANALYTICAL")

    # ---------------------------------------------------
    # Initialize the database and Load the Root node
    # ---------------------------------------------------
    async with lock.registry.initialization():
        first_time_initialization = len(await Root.get_list(db=db)) == 0
        log.debug("Checking Root Node")
        await initialize_registry(db=db, initialize=True)

        if add_database_indexes:
            await add_indexes(db=db)

    # ---------------------------------------------------
    # Load all schema in the database into the registry
    #  ... Unless the schema has been initialized already
    # ---------------------------------------------------
    if not registry.schema_has_been_initialized():
        registry.schema = SchemaManager()
        schema = SchemaRoot(**internal_schema)
        registry.schema.register_schema(schema=schema)

        # Import the default branch
        default_branch: Branch = registry.get_branch_from_registry(branch=registry.default_branch)
        hash_in_db = default_branch.active_schema_hash.main
        schema_default_branch = await registry.schema.load_schema_from_db(db=db, branch=default_branch)
        registry.schema.set_schema_branch(name=default_branch.name, schema=schema_default_branch)

        if default_branch.update_schema_hash():
            log.warning(
                "New schema detected after pulling the schema from the db",
                hash_current=hash_in_db,
                hash_new=default_branch.active_schema_hash.main,
                branch=default_branch.name,
            )
            await default_branch.save(db=db)

        for branch in list(registry.branch.values()):
            if branch.name in [default_branch.name, GLOBAL_BRANCH_NAME]:
                continue

            hash_in_db = branch.active_schema_hash.main
            log.info("Importing schema", branch=branch.name)
            await registry.schema.load_schema(db=db, branch=branch)

            if branch.update_schema_hash():
                log.warning(
                    f"New schema detected after pulling the schema from the db :"
                    f" {hash_in_db!r} >> {branch.active_schema_hash.main!r}",
                    branch=branch.name,
                )
                await branch.save(db=db)

    default_branch = registry.get_branch_from_registry(branch=registry.default_branch)
    schema_branch = registry.schema.get_schema_branch(name=default_branch.name)
    gqlm = graphql_registry.get_manager_for_branch(branch=default_branch, schema_branch=schema_branch)
    gqlm.get_graphql_schema(
        include_query=True,
        include_mutation=True,
        include_subscription=True,
        include_types=True,
    )

    # ---------------------------------------------------
    # Load Default Namespace
    # ---------------------------------------------------
    ip_namespace = await get_default_ipnamespace(db=db)
    if ip_namespace:
        registry.default_ipnamespace = ip_namespace.id
    return first_time_initialization


async def create_root_node(db: InfrahubDatabase) -> Root:
    root = Root(graph_version=GRAPH_VERSION, default_branch=config.SETTINGS.initial.default_branch)
    await root.save(db=db)
    log.info(f"Generated instance ID : {root.uuid} (v{GRAPH_VERSION})")

    registry.id = root.id
    registry.default_branch = root.default_branch

    return root


async def create_default_branch(db: InfrahubDatabase) -> Branch:
    branch = Branch(
        name=registry.default_branch,
        status=BranchStatus.OPEN,
        description="Default Branch",
        hierarchy_level=1,
        is_default=True,
        sync_with_git=True,
    )
    await branch.save(db=db)
    registry.branch[branch.name] = branch

    log.info("Created default branch", branch=branch.name)

    return branch


async def create_global_branch(db: InfrahubDatabase) -> Branch:
    branch = Branch(
        name=GLOBAL_BRANCH_NAME,
        status=BranchStatus.OPEN,
        description="Global Branch",
        hierarchy_level=1,
        is_global=True,
        sync_with_git=False,
    )
    await branch.save(db=db)
    registry.branch[branch.name] = branch

    log.info("Created global branch", branch=branch.name)

    return branch


async def create_branch(
    branch_name: str, db: InfrahubDatabase, description: str = "", isolated: bool = True, at: str | None = None
) -> Branch:
    """Create a new Branch, currently all the branches are based on Main

    Because all branches are based on main, the hierarchy_level of hardcoded to 2."""
    description = description or f"Branch {branch_name}"
    branch = Branch(
        name=branch_name,
        status=BranchStatus.OPEN,
        hierarchy_level=2,
        description=description,
        is_default=False,
        sync_with_git=False,
        created_at=at,
        branched_from=at,
        is_isolated=isolated,
    )

    origin_schema = registry.schema.get_schema_branch(name=branch.origin_branch)
    new_schema = origin_schema.duplicate(name=branch.name)
    registry.schema.set_schema_branch(name=branch.name, schema=new_schema)

    branch.update_schema_hash()
    await branch.save(db=db)
    registry.branch[branch.name] = branch

    log.info("Created branch", branch=branch.name)

    return branch


async def create_account(
    db: InfrahubDatabase,
    name: str = "admin",
    password: str | None = None,
    token_value: str | None = None,
) -> CoreAccount:
    token_schema = db.schema.get_node_schema(name=InfrahubKind.ACCOUNTTOKEN)
    obj = await Node.init(db=db, schema=CoreAccount)
    await obj.new(db=db, name=name, account_type="User", password=password)
    await obj.save(db=db)
    log.info(f"Created Account: {name}", account_name=name)

    if token_value:
        token = await Node.init(db=db, schema=token_schema)
        await token.new(db=db, token=token_value, name="Created automatically", account=obj)
        await token.save(db=db)

    return obj


async def create_ipam_namespace(
    db: InfrahubDatabase,
    name: str = DEFAULT_IP_NAMESPACE,
    description: str = "Used to provide a default space of IP resources",
) -> Node:
    obj = await Node.init(db=db, schema=InfrahubKind.NAMESPACE)
    await obj.new(db=db, name=name, description=description, default=True)
    await obj.save(db=db)
    log.info(f"Created IPAM Namespace: {name}")

    return obj


async def create_super_administrator_role(db: InfrahubDatabase) -> CoreAccountRole:
    permission = await Node.init(db=db, schema=InfrahubKind.GLOBALPERMISSION)
    await permission.new(
        db=db,
        action=GlobalPermissions.SUPER_ADMIN.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to do anything",
    )
    await permission.save(db=db)
    log.info(f"Created global permission: {GlobalPermissions.SUPER_ADMIN}")

    role_name = "Super Administrator"
    role = await Node.init(db=db, schema=CoreAccountRole)
    await role.new(db=db, name=role_name, permissions=[permission])
    await role.save(db=db)
    log.info(f"Created account role: {role_name}")

    return role


async def create_default_role(db: InfrahubDatabase) -> CoreAccountRole:
    repo_permission = await Node.init(db=db, schema=InfrahubKind.GLOBALPERMISSION)
    await repo_permission.new(
        db=db,
        action=GlobalPermissions.MANAGE_REPOSITORIES.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to manage repositories",
    )
    await repo_permission.save(db=db)

    schema_permission = await Node.init(db=db, schema=InfrahubKind.GLOBALPERMISSION)
    await schema_permission.new(
        db=db,
        action=GlobalPermissions.MANAGE_SCHEMA.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to manage the schema",
    )
    await schema_permission.save(db=db)

    proposed_change_permission = await Node.init(db=db, schema=InfrahubKind.GLOBALPERMISSION)
    await proposed_change_permission.new(
        db=db,
        action=GlobalPermissions.MERGE_PROPOSED_CHANGE.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to merge proposed changes",
    )
    await proposed_change_permission.save(db=db)

    # Other permissions, created to keep references of them from the start
    for permission_action in (
        GlobalPermissions.EDIT_DEFAULT_BRANCH,
        GlobalPermissions.MANAGE_ACCOUNTS,
        GlobalPermissions.MANAGE_PERMISSIONS,
        GlobalPermissions.MERGE_BRANCH,
    ):
        await get_or_create_global_permission(db=db, permission=permission_action)

    view_permission = await Node.init(db=db, schema=InfrahubKind.OBJECTPERMISSION)
    await view_permission.new(
        db=db,
        name="*",
        namespace="*",
        action=PermissionAction.VIEW.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to view any object in any branch",
    )
    await view_permission.save(db=db)

    modify_permission = await Node.init(db=db, schema=InfrahubKind.OBJECTPERMISSION)
    await modify_permission.new(
        db=db,
        name="*",
        namespace="*",
        action=PermissionAction.ANY.value,
        decision=PermissionDecision.ALLOW_OTHER.value,
        description="Allow a user to change data in non-default branches",
    )
    await modify_permission.save(db=db)

    role_name = "General Access"
    role = await Node.init(db=db, schema=CoreAccountRole)
    await role.new(
        db=db,
        name=role_name,
        permissions=[
            repo_permission,
            schema_permission,
            proposed_change_permission,
            view_permission,
            modify_permission,
        ],
    )
    await role.save(db=db)
    log.info(f"Created account role: {role_name}")

    return role


async def create_proposed_change_reviewer_role(db: InfrahubDatabase) -> CoreAccountRole:
    edit_default_branch_permission = await get_or_create_global_permission(
        db=db, permission=GlobalPermissions.EDIT_DEFAULT_BRANCH
    )
    reviewer_permission = await get_or_create_global_permission(
        db=db, permission=GlobalPermissions.REVIEW_PROPOSED_CHANGE
    )

    proposed_change_update_permission = await Node.init(db=db, schema=InfrahubKind.OBJECTPERMISSION)
    await proposed_change_update_permission.new(
        db=db,
        name="ProposedChange",
        namespace="Core",
        action=PermissionAction.UPDATE.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description="Allow a user to update proposed changes",
    )
    await proposed_change_update_permission.save(db=db)

    role_name = "Proposed Change Reviewer"
    role = await Node.init(db=db, schema=CoreAccountRole)
    await role.new(
        db=db,
        name=role_name,
        permissions=[edit_default_branch_permission, reviewer_permission, proposed_change_update_permission],
    )
    await role.save(db=db)
    log.info(f"Created account role: {role_name}")

    return role


async def create_anonymous_role(db: InfrahubDatabase) -> CoreAccountRole:
    deny_permission = await Node.init(db=db, schema=InfrahubKind.OBJECTPERMISSION)
    await deny_permission.new(
        db=db, name="*", namespace="*", action=PermissionAction.ANY.value, decision=PermissionDecision.DENY.value
    )
    await deny_permission.save(db=db)

    view_permission = await NodeManager.get_one_by_hfid(
        db=db,
        kind=InfrahubKind.OBJECTPERMISSION,
        hfid=["*", "*", PermissionAction.VIEW.value, str(PermissionDecision.ALLOW_ALL.value)],
    )

    role = await Node.init(db=db, schema=CoreAccountRole)
    await role.new(
        db=db, name=config.SETTINGS.main.anonymous_access_role, permissions=[deny_permission, view_permission]
    )
    await role.save(db=db)
    log.info(f"Created anonymous account role: {config.SETTINGS.main.anonymous_access_role}")

    return role


async def create_accounts_group(
    db: InfrahubDatabase, name: str, roles: Sequence[CoreAccountRole], accounts: Sequence[CoreAccount]
) -> CoreAccountGroup:
    group = await Node.init(db=db, schema=CoreAccountGroup)
    await group.new(db=db, name=name, roles=list(roles))
    await group.save(db=db)
    log.info(f"Created account group: {name}")

    for account in accounts:
        await group.members.add(db=db, data=account)  # type: ignore[arg-type]
        await group.members.save(db=db)
        log.info(f"Assigned account group: {name} to {account.name.value}")

    return group


async def create_default_account_groups(
    db: InfrahubDatabase,
    admin_accounts: Sequence[CoreAccount] | None = None,
    accounts: Sequence[CoreAccount] | None = None,
) -> None:
    administrator_role = await create_super_administrator_role(db=db)
    await create_accounts_group(
        db=db, name="Super Administrators", roles=[administrator_role], accounts=admin_accounts or []
    )

    default_role = await create_default_role(db=db)
    proposed_change_reviewer_role = await create_proposed_change_reviewer_role(db=db)
    await create_accounts_group(
        db=db, name="Infrahub Users", roles=[default_role, proposed_change_reviewer_role], accounts=accounts or []
    )


async def first_time_initialization(db: InfrahubDatabase) -> None:
    # --------------------------------------------------
    # Create the default Branch
    # --------------------------------------------------
    await create_root_node(db=db)
    default_branch = await create_default_branch(db=db)
    await create_global_branch(db=db)

    # --------------------------------------------------
    # Load the internal and core schema in the database
    # --------------------------------------------------
    registry.schema = SchemaManager()
    schema = SchemaRoot(**internal_schema)
    schema_branch = registry.schema.register_schema(schema=schema, branch=default_branch.name)
    schema_branch.load_schema(schema=SchemaRoot(**core_models))
    schema_branch.process()
    await registry.schema.load_schema_to_db(schema=schema_branch, branch=default_branch, db=db, at=Timestamp())
    registry.schema.set_schema_branch(name=default_branch.name, schema=schema_branch)
    default_branch.update_schema_hash()
    await default_branch.save(db=db)
    log.info("Created the Schema in the database", hash=default_branch.active_schema_hash.main)

    # --------------------------------------------------
    # Create Default Menu
    # --------------------------------------------------
    await create_default_menu(db=db)

    # --------------------------------------------------
    # Create Default Users and Groups
    # --------------------------------------------------
    admin_accounts: list[CoreAccount] = []
    admin_accounts.append(
        await create_account(
            db=db,
            name="admin",
            password=config.SETTINGS.initial.admin_password,
            token_value=config.SETTINGS.initial.admin_token,
        )
    )

    if config.SETTINGS.initial.create_agent_user:
        password = config.SETTINGS.initial.agent_password or str(uuid4())

        admin_accounts.append(
            await create_account(
                db=db, name="agent", password=password, token_value=config.SETTINGS.initial.agent_token
            )
        )

    # --------------------------------------------------
    # Create default account roles, groups and permissions
    # --------------------------------------------------
    await create_default_account_groups(db=db, admin_accounts=admin_accounts)

    if config.SETTINGS.main.allow_anonymous_access:
        await create_anonymous_role(db=db)

    # --------------------------------------------------
    # Create Default IPAM Namespace
    # --------------------------------------------------
    await create_ipam_namespace(db=db)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import (
    OBJECT_TEMPLATE_NAME_ATTR,
    OBJECT_TEMPLATE_RELATIONSHIP_NAME,
    RelationshipCardinality,
    RelationshipKind,
)
from infrahub.core.constants.schema import RESOURCE_POOL_REL_SUFFIX

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.protocols import CoreObjectTemplate
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.database import InfrahubDatabase
    from infrahub.pools.allocator import PoolAllocator


class NodeTemplateApplier:
    """Applies a template to produce field data for a new node."""

    def __init__(self, db: InfrahubDatabase, branch: Branch, pool_allocator: PoolAllocator) -> None:
        self.db = db
        self.branch = branch
        self.pool_allocator = pool_allocator

    async def apply(
        self,
        template: CoreObjectTemplate,
        target_schema: MainSchemaTypes,
        target_id: str,
        user_fields: dict[str, Any],
    ) -> dict[str, Any]:
        """Apply template and return merged fields. User fields take precedence."""
        fields = dict(user_fields)
        await self._apply_attributes(template=template, fields=fields)
        await self._apply_relationships(
            template=template, target_schema=target_schema, target_id=target_id, fields=fields
        )
        return fields

    async def _apply_attributes(self, template: CoreObjectTemplate, fields: dict[str, Any]) -> None:
        for attr_name in template.get_schema().attribute_names:
            if attr_name in fields or attr_name == OBJECT_TEMPLATE_NAME_ATTR:
                continue

            attr = template.get_attribute(name=attr_name)
            if attr.value is None:
                continue

            # source_id is dynamically set by NodePropertyMixin._init_node_property_mixin hence the type hint ignore
            field_data: dict[str, Any] = {"value": attr.value, "source": attr.source_id or template.id}  # type: ignore[attr-defined]
            if attr.is_from_profile:
                field_data["is_from_profile"] = True
            fields[attr_name] = field_data

    async def _apply_relationships(
        self, template: CoreObjectTemplate, target_schema: MainSchemaTypes, target_id: str, fields: dict[str, Any]
    ) -> None:
        template_schema = template.get_schema()
        for rel_name in template_schema.relationship_names:
            rel_schema = template_schema.get_relationship(name=rel_name)

            if rel_name.endswith(RESOURCE_POOL_REL_SUFFIX):
                await self._handle_pool_relationship(
                    template=template,
                    pool_rel_name=rel_name,
                    target_schema=target_schema,
                    target_id=target_id,
                    fields=fields,
                )
                continue

            if rel_name in fields:
                continue

            if rel_schema.kind not in [RelationshipKind.ATTRIBUTE, RelationshipKind.GENERIC, RelationshipKind.PROFILE]:
                continue

            if rel_name == OBJECT_TEMPLATE_RELATIONSHIP_NAME:
                continue

            relationship = template.get_relationship(name=rel_name)
            if rel_schema.cardinality == RelationshipCardinality.ONE:
                if peer := await relationship.get_peer(db=self.db):
                    fields[rel_name] = {"id": peer.id}
            elif peers := await relationship.get_peers(db=self.db):
                fields[rel_name] = [{"id": peer_id} for peer_id in peers]

    async def _handle_pool_relationship(
        self,
        template: CoreObjectTemplate,
        pool_rel_name: str,
        target_schema: MainSchemaTypes,
        target_id: str,
        fields: dict[str, Any],
    ) -> None:
        """Allocate from pool and set the original relationship or attribute."""
        original_name = pool_rel_name.removesuffix(RESOURCE_POOL_REL_SUFFIX)

        if original_name in fields:
            return

        pool_relationship = template.get_relationship(name=pool_rel_name)

        if original_name in target_schema.attribute_names:
            # Number pool: allocate a value for the attribute
            allocated_value = await self.pool_allocator.allocate_for_attribute(
                pool_relationship=pool_relationship,
                target_schema=target_schema,
                attribute_name=original_name,
                identifier=target_id,
            )
            if allocated_value is not None:
                pool_node = await pool_relationship.get_peer(db=self.db, raise_on_error=True)
                fields[original_name] = {"value": allocated_value, "source": pool_node.id}
            return
        # IP pool: allocate a node for the relationship
        allocated = await self.pool_allocator.allocate_for_relationship(
            pool_relationship=pool_relationship, identifier=target_id
        )
        if allocated:
            pool_node = await pool_relationship.get_peer(db=self.db, raise_on_error=True)
            fields[original_name] = {"peer": allocated, "_relation__source": pool_node.id}



from __future__ import annotations

import ast
from typing import TYPE_CHECKING

from ariadne_codegen.plugins.base import Plugin

if TYPE_CHECKING:
    from graphql import ExecutableDefinitionNode


class FutureAnnotationPlugin(Plugin):
    @staticmethod
    def insert_future_annotation(module: ast.Module) -> ast.Module:
        # First check if the future annotation is already present
        for item in module.body:
            if (
                isinstance(item, ast.ImportFrom)
                and item.module == "__future__"
                and any(alias.name == "annotations" for alias in item.names)
            ):
                return module

        module.body.insert(0, ast.ImportFrom(module="__future__", names=[ast.alias(name="annotations")], level=0))
        return module

    def generate_result_types_module(
        self,
        module: ast.Module,
        operation_definition: ExecutableDefinitionNode,  # noqa: ARG002
    ) -> ast.Module:
        return self.insert_future_annotation(module)


class StandardTypeHintPlugin(Plugin):
    @classmethod
    def replace_list_in_subscript(cls, subscript: ast.Subscript) -> ast.Subscript:
        if isinstance(subscript.value, ast.Name) and subscript.value.id == "List":
            subscript.value.id = "list"
        if isinstance(subscript.slice, ast.Subscript):
            subscript.slice = cls.replace_list_in_subscript(subscript.slice)

        return subscript

    @classmethod
    def replace_list_annotations(cls, module: ast.Module) -> ast.Module:
        for item in module.body:
            if not isinstance(item, ast.ClassDef):
                continue

            # replace List with list in the annotations when list is used as a type
            for class_item in item.body:
                if not isinstance(class_item, ast.AnnAssign):
                    continue
                if isinstance(class_item.annotation, ast.Subscript):
                    class_item.annotation = cls.replace_list_in_subscript(class_item.annotation)

        return module

    def generate_result_types_module(
        self,
        module: ast.Module,
        operation_definition: ExecutableDefinitionNode,  # noqa: ARG002
    ) -> ast.Module:
        module = FutureAnnotationPlugin.insert_future_annotation(module)
        return self.replace_list_annotations(module)


class PydanticBaseModelPlugin(Plugin):
    @staticmethod
    def find_base_model_index(module: ast.Module) -> int:
        for idx, item in enumerate(module.body):
            if isinstance(item, ast.ImportFrom) and item.module == "base_model":
                return idx
        raise ValueError("BaseModel not found in module")

    @classmethod
    def replace_base_model_import(cls, module: ast.Module) -> ast.Module:
        base_model_index = cls.find_base_model_index(module)
        module.body[base_model_index] = ast.ImportFrom(module="pydantic", names=[ast.alias(name="BaseModel")], level=0)
        return module

    def generate_result_types_module(
        self,
        module: ast.Module,
        operation_definition: ExecutableDefinitionNode,  # noqa: ARG002
    ) -> ast.Module:
        return self.replace_base_model_import(module)



from infrahub import lock


class MergeLocker:
    lock_namespace = "merge"

    def __init__(self) -> None:
        self.lock_registry = lock.registry

    def acquire_global_lock(self) -> lock.InfrahubLock:
        return self.lock_registry.get(name="all_branches", namespace=self.lock_namespace)



from typing import Any

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


class EnrichedDiffDeleteQuery(Query):
    name = "enriched_diff_delete"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        enriched_diff_root_uuids: list[str] | None = None,
        include_frozen: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.enriched_diff_root_uuids = enriched_diff_root_uuids
        self.include_frozen = include_frozen

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        diff_filters = []
        self.params = {}
        if self.enriched_diff_root_uuids:
            self.params["diff_root_uuids"] = self.enriched_diff_root_uuids
            diff_filters.append("d_root.uuid IN $diff_root_uuids")
        if not self.include_frozen:
            diff_filters.append("(d_root.is_frozen IS NULL OR d_root.is_frozen <> TRUE)")

        diff_filter = ""
        if diff_filters:
            diff_filter = "WHERE " + " AND ".join(diff_filters)

        query = """
MATCH (d_root:DiffRoot)
%(diff_filter)s
OPTIONAL MATCH (d_root)-[*]->(diff_thing:DiffRoot|DiffNode|DiffAttribute|DiffRelationship|DiffRelationshipElement|DiffProperty|DiffConflict)
WITH DISTINCT d_root, diff_thing
ORDER BY elementId(diff_thing)
CALL (diff_thing) {
    DETACH DELETE diff_thing
} IN TRANSACTIONS
DETACH DELETE d_root
        """ % {"diff_filter": diff_filter}
        self.add_to_query(query=query)



from __future__ import annotations

import enum
from enum import Enum
from typing import TYPE_CHECKING, Any, Self

from pydantic import Field, PrivateAttr, ValidationInfo, field_validator, model_validator

from infrahub import config
from infrahub.core.constants.schema import UpdateSupport
from infrahub.core.enums import generate_python_enum
from infrahub.core.query.attribute import default_attribute_query_filter
from infrahub.exceptions import InitializationError
from infrahub.types import ATTRIBUTE_KIND_LABELS, ATTRIBUTE_TYPES

from .attribute_parameters import (
    AttributeParameters,
    ListAttributeParameters,
    NumberAttributeParameters,
    NumberPoolParameters,
    TextAttributeParameters,
    get_attribute_parameters_class_for_kind,
)
from .generated.attribute_schema import GeneratedAttributeSchema

if TYPE_CHECKING:
    from infrahub.core.attribute import BaseAttribute
    from infrahub.core.branch import Branch
    from infrahub.core.constants import BranchSupportType
    from infrahub.core.query import QueryElement
    from infrahub.database import InfrahubDatabase


def get_attribute_schema_class_for_kind(kind: str) -> type[AttributeSchema]:
    return attribute_schema_class_by_kind.get(kind, AttributeSchema)


class AttributeSchema(GeneratedAttributeSchema):
    _sort_by: list[str] = ["name"]
    _enum_class: type[enum.Enum] | None = None
    # Stores the source generic's attribute ID for inherited attributes
    # Used for rename detection in diffs while keeping id=None for inherited attrs
    _source_attribute_id: str | None = PrivateAttr(default=None)

    @property
    def source_attribute_id(self) -> str | None:
        return self._source_attribute_id

    @source_attribute_id.setter
    def source_attribute_id(self, value: str | None) -> None:
        self._source_attribute_id = value

    @classmethod
    def model_json_schema(cls, *args: Any, **kwargs: Any) -> dict[str, Any]:
        schema = super().model_json_schema(*args, **kwargs)

        # Build conditional schema based on attribute_schema_class_by_kind mapping
        # This override allows people using the Yaml language server to get the correct mappings
        # for the parameters when selecting the appropriate kind
        schema["allOf"] = []
        for kind, schema_class in attribute_schema_class_by_kind.items():
            schema["allOf"].append(
                {
                    "if": {"properties": {"kind": {"const": kind}}},
                    "then": {"properties": {"parameters": {"$ref": f"#/definitions/{schema_class.__name__}"}}},
                }
            )

        return schema

    @property
    def is_attribute(self) -> bool:
        return True

    @property
    def is_relationship(self) -> bool:
        return False

    @property
    def is_deprecated(self) -> bool:
        return bool(self.deprecation)

    @property
    def support_profiles(self) -> bool:
        return self.read_only is False and self.unique is False

    @property
    def support_templates(self) -> bool:
        return self.read_only is False and self.unique is False

    def get_id(self) -> str:
        if self.id is None:
            raise InitializationError("The attribute schema has not been saved yet and doesn't have an id")
        return self.id

    def to_dict(self) -> dict:
        data = self.model_dump(exclude_unset=True, exclude_none=True)
        for field_name, value in data.items():
            if isinstance(value, Enum):
                data[field_name] = value.value
        return data

    @field_validator("kind")
    @classmethod
    def kind_options(cls, v: str) -> str:
        if v not in ATTRIBUTE_KIND_LABELS:
            raise ValueError(f"Only valid Attribute Kind are : {ATTRIBUTE_KIND_LABELS} ")
        return v

    @model_validator(mode="before")
    @classmethod
    def validate_dropdown_choices(cls, values: Any) -> Any:
        """Validate that choices are defined for a dropdown but not for other kinds."""
        if isinstance(values, dict):
            kind = values.get("kind")
            choices = values.get("choices")
        elif isinstance(values, AttributeSchema):
            kind = values.kind
            choices = values.choices
        else:
            return values
        if kind != "Dropdown" and choices:
            raise ValueError(f"Can only specify 'choices' for kind=Dropdown: {kind}")

        if kind == "Dropdown" and not choices:
            raise ValueError("The property 'choices' is required for kind=Dropdown")

        return values

    @field_validator("parameters", mode="before")
    @classmethod
    def set_parameters_type(cls, value: Any, info: ValidationInfo) -> Any:
        """Override parameters class if using base AttributeParameters class and should be using a subclass.

        This validator handles parameter type conversion when an attribute's kind changes.
        Fields from the source that don't exist in the target are silently dropped.
        Fields with the same name in both classes are preserved.
        """
        kind = info.data["kind"]
        expected_parameters_class = get_attribute_parameters_class_for_kind(kind=kind)
        if value is None:
            return expected_parameters_class()
        if not isinstance(value, expected_parameters_class) and isinstance(value, AttributeParameters):
            return expected_parameters_class.convert_from(value)
        if isinstance(value, dict):
            return expected_parameters_class.convert_from_dict(source_data=value)
        return value

    @model_validator(mode="after")
    def validate_parameters(self) -> Self:
        if isinstance(self.parameters, NumberPoolParameters) and not self.kind == "NumberPool":
            raise ValueError(f"NumberPoolParameters can't be used as parameters for {self.kind}")

        if isinstance(self.parameters, TextAttributeParameters) and self.kind not in ["Text", "TextArea"]:
            raise ValueError(f"TextAttributeParameters can't be used as parameters for {self.kind}")

        if isinstance(self.parameters, ListAttributeParameters) and self.kind != "List":
            raise ValueError(f"ListAttributeParameters can't be used as parameters for {self.kind}")

        return self

    def get_class(self) -> type[BaseAttribute]:
        return ATTRIBUTE_TYPES[self.kind].get_infrahub_class()

    @property
    def uses_enum_class(self) -> bool:
        return bool(self.enum) and config.SETTINGS.experimental_features.graphql_enums

    def get_branch(self) -> BranchSupportType:
        if not self.branch:
            raise ValueError("branch hasn't been defined yet")
        return self.branch

    def get_enum_class(self) -> type[enum.Enum]:
        if not self.enum:
            raise ValueError(f"{self.name} is not an Enum")
        if not self._enum_class:
            self._enum_class = generate_python_enum(name=f"{self.name.title()}Enum", options=self.enum)
        return self._enum_class

    def convert_value_to_enum(self, value: Any) -> enum.Enum | None:
        if isinstance(value, enum.Enum) or value is None:
            return value
        enum_class = self.get_enum_class()
        return enum_class(value)

    def convert_enum_to_value(self, data: Any) -> Any:
        if isinstance(data, list):
            value = [self.convert_enum_to_value(element) for element in data]
            return [element.value if isinstance(element, enum.Enum) else element for element in value]
        if isinstance(data, enum.Enum):
            return data.value
        return data

    def update_from_generic(self, other: AttributeSchema) -> None:
        fields_to_exclude = ("id", "order_weight", "branch", "inherited")
        for name in self.__class__.model_fields:
            if name in fields_to_exclude:
                continue
            if getattr(self, name) != getattr(other, name):
                setattr(self, name, getattr(other, name))

    def to_node(self) -> dict[str, Any]:
        fields_to_exclude = {"id", "state", "filters"}
        fields_to_json = {"computed_attribute", "parameters"}
        data = self.model_dump(exclude=fields_to_exclude | fields_to_json)

        for field_name in fields_to_json:
            if field := getattr(self, field_name):
                data[field_name] = {"value": field.model_dump()}
            else:
                data[field_name] = None

        return data

    def get_regex(self) -> str | None:
        return self.regex

    def get_min_length(self) -> int | None:
        return self.min_length

    def get_max_length(self) -> int | None:
        return self.max_length

    async def get_query_filter(
        self,
        name: str,
        filter_name: str,
        branch: Branch | None = None,
        filter_value: str | int | bool | list | None = None,
        include_match: bool = True,
        param_prefix: str | None = None,
        db: InfrahubDatabase | None = None,
        partial_match: bool = False,
    ) -> tuple[list[QueryElement], dict[str, Any], list[str]]:
        if self.enum:
            filter_value = self.convert_enum_to_value(filter_value)

        return await default_attribute_query_filter(
            name=name,
            attribute_kind=self.kind,
            filter_name=filter_name,
            branch=branch,
            filter_value=filter_value,
            include_match=include_match,
            param_prefix=param_prefix,
            db=db,
            partial_match=partial_match,
        )


class NumberPoolSchema(AttributeSchema):
    parameters: NumberPoolParameters = Field(
        default_factory=NumberPoolParameters,
        description="Extra parameters specific to NumberPool attributes",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )


class TextAttributeSchema(AttributeSchema):
    parameters: TextAttributeParameters = Field(
        default_factory=TextAttributeParameters,
        description="Extra parameters specific to text attributes",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )

    def get_regex(self) -> str | None:
        return self.parameters.regex

    def get_min_length(self) -> int | None:
        return self.parameters.min_length

    def get_max_length(self) -> int | None:
        return self.parameters.max_length


class NumberAttributeSchema(AttributeSchema):
    parameters: NumberAttributeParameters = Field(
        default_factory=NumberAttributeParameters,
        description="Extra parameters specific to number attributes",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )


class ListAttributeSchema(AttributeSchema):
    """Schema for List attributes with regex validation support.

    Note: Only regex validation is supported for List attributes.
    Unlike Text/TextArea attributes, min_length and max_length are not supported.
    """

    parameters: ListAttributeParameters = Field(
        default_factory=ListAttributeParameters,
        description="Extra parameters specific to list attributes",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )

    def get_regex(self) -> str | None:
        return self.parameters.regex


attribute_schema_class_by_kind: dict[str, type[AttributeSchema]] = {
    "NumberPool": NumberPoolSchema,
    "Text": TextAttributeSchema,
    "TextArea": TextAttributeSchema,
    "List": ListAttributeSchema,
    "Number": NumberAttributeSchema,
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from graphene import Boolean, InputObjectType, List, Mutation, NonNull, String

from infrahub.core.account import ObjectPermission
from infrahub.core.constants import GlobalPermissions, PermissionAction, PermissionDecision
from infrahub.core.manager import NodeManager
from infrahub.core.registry import registry
from infrahub.database import retry_db_transaction
from infrahub.events import EventMeta
from infrahub.events.node_action import NodeUpdatedEvent
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.types.context import ContextInput
from infrahub.log import get_log_data
from infrahub.permissions import define_global_permission_from_branch
from infrahub.worker import WORKER_IDENTITY

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class InfrahubHFIDUpdateInput(InputObjectType):
    id = String(required=True)
    kind = String(required=True)
    value = List(NonNull(String), required=True)


class UpdateHFID(Mutation):
    class Arguments:
        data = InfrahubHFIDUpdateInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="update_hfid")
    async def mutate(
        cls,
        _: dict,
        info: GraphQLResolveInfo,
        data: InfrahubHFIDUpdateInput,
        context: ContextInput | None = None,
    ) -> UpdateHFID:
        graphql_context: GraphqlContext = info.context
        node_schema = registry.schema.get_node_schema(
            name=str(data.kind), branch=graphql_context.branch.name, duplicate=False
        )
        if node_schema.human_friendly_id is None:
            raise ValidationError(
                input_value=f"{node_schema.kind}.human_friendly_id has not been defined for this kind."
            )

        updated_hfid = cast("list[str]", data.value)

        if len(node_schema.human_friendly_id) != len(updated_hfid):
            raise ValidationError(
                input_value=f"{node_schema.kind}.human_friendly_id requires {len(node_schema.human_friendly_id)} parts data has {len(updated_hfid)}"
            )

        graphql_context.active_permissions.raise_for_permissions(
            permissions=[
                define_global_permission_from_branch(
                    permission=GlobalPermissions.UPDATE_OBJECT_HFID_DISPLAY_LABEL,
                    branch_name=graphql_context.branch.name,
                ),
                ObjectPermission(
                    namespace=node_schema.namespace,
                    name=node_schema.name,
                    action=PermissionAction.UPDATE.value,
                    decision=PermissionDecision.ALLOW_DEFAULT.value
                    if graphql_context.branch.name == registry.default_branch
                    else PermissionDecision.ALLOW_OTHER.value,
                ),
            ]
        )
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        if not (
            target_node := await NodeManager.get_one(
                db=graphql_context.db,
                kind=node_schema.kind,
                id=str(data.id),
                branch=graphql_context.branch,
                fields={"human_friendly_id": None},
            )
        ):
            raise NodeNotFoundError(
                node_type=node_schema.kind,
                identifier=str(data.id),
                message="The targeted node was not found in the database",
            )

        existing = await target_node.get_hfid(db=graphql_context.db) if target_node.has_human_friendly_id() else None
        if updated_hfid != existing:
            await target_node.set_human_friendly_id(value=updated_hfid)

            async with graphql_context.db.start_transaction() as dbt:
                await target_node.save(db=dbt, fields=["human_friendly_id"], user_id=graphql_context.assigned_user_id)

            log_data = get_log_data()
            request_id = log_data.get("request_id", "")

            event = NodeUpdatedEvent(
                kind=node_schema.kind,
                node_id=target_node.get_id(),
                changelog=target_node.node_changelog,
                fields=["human_friendly_id"],
                meta=EventMeta(
                    context=graphql_context.get_context(),
                    initiator_id=WORKER_IDENTITY,
                    request_id=request_id,
                    account_id=graphql_context.active_account_session.account_id,
                    branch=graphql_context.branch,
                ),
            )
            await graphql_context.active_service.event.send(event=event)

        result: dict[str, Any] = {"ok": True}

        return cls(**result)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import GraphMigration, MigrationInput, MigrationResult

from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class CleanupOrphanedNodesQuery(Query):
    """
    Clean up orphaned Node vertices (no IS_PART_OF edge to Root) and their linked
    Attributes and Relationships.
    """

    name = "cleanup_orphaned_nodes"
    type = QueryType.WRITE
    insert_return = False
    raise_error_if_empty = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
// Delete attributes of orphaned nodes
MATCH (n:Node)
WHERE NOT exists((n)-[:IS_PART_OF]->(:Root))
OPTIONAL MATCH (n)-[:HAS_ATTRIBUTE]->(attr:Attribute)
WITH DISTINCT attr
CALL (attr) {
    DETACH DELETE attr
} IN TRANSACTIONS

// reduce the results to a single row
WITH 1 AS one
LIMIT 1

// Delete relationships that will have < 2 Node peers after orphaned node removal
OPTIONAL MATCH (orphan:Node)-[:IS_RELATED]-(rel:Relationship)
WHERE NOT exists((orphan)-[:IS_PART_OF]->(:Root))
WITH DISTINCT rel
CALL (rel) {
    OPTIONAL MATCH (rel)-[:IS_RELATED]-(peer:Node)
    WHERE exists((peer)-[:IS_PART_OF]->(:Root))
    WITH rel, count(peer) AS remaining_peers
    WHERE remaining_peers < 2
    DETACH DELETE rel
} IN TRANSACTIONS

// reduce the results to a single row
WITH 1 AS one
LIMIT 1

// Delete the orphaned nodes
MATCH (n:Node)
WHERE NOT exists((n)-[:IS_PART_OF]->(:Root))
CALL (n) {
    DETACH DELETE n
} IN TRANSACTIONS
        """
        self.add_to_query(query)


class Migration054(GraphMigration):
    """
    Clean up orphaned Node vertices that have no IS_PART_OF edge to Root.

    This can happen when a branch-aware node is deleted during branch deletion,
    but its branch-agnostic attributes or relationships are not properly cleaned up.

    The migration:
    1. DETACH DELETEs Attributes linked to orphaned nodes
    2. DETACH DELETEs Relationships that would have < 2 Node peers after orphaned node removal
    3. DETACH DELETEs the orphaned nodes themselves
    """

    name: str = "054_cleanup_orphaned_nodes"
    minimum_version: int = 53
    queries: Sequence[type[Query]] = [CleanupOrphanedNodesQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        # Override parent class to skip transaction in case there are many nodes to delete
        return await self.do_execute(migration_input=migration_input)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.branch.enums import BranchStatus
from infrahub.core.branch.filters import BranchListFilters
from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.core.query import Query, QueryType
from infrahub.core.query.standard_node import StandardNodeGetListQuery
from infrahub.core.timestamp import Timestamp

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class DeleteBranchRelationshipsQuery(Query):
    name: str = "delete_branch_relationships"
    insert_return: bool = False

    type: QueryType = QueryType.WRITE

    def __init__(self, branch_name: str, **kwargs: Any) -> None:
        self.branch_name = branch_name
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        query = """
// --------------
// for every Node that only exists on this branch (it's about to be deleted),
// find any agnostic relationships or attributes connected to the Node and delete them
// --------------
OPTIONAL MATCH (:Root)<-[e:IS_PART_OF {status: "active"}]-(n:Node)
WHERE e.branch = $branch_name
// does the node only exist on this branch?
CALL (n) {
    OPTIONAL MATCH (n)-[ipo:IS_PART_OF {status: "active"}]->(:Root)
    WHERE ipo.branch <> $branch_name
    LIMIT 1
    RETURN ipo IS NOT NULL AS node_exists_on_other_branch
}
// if so, delete any linked agnostic relationships or attributes
CALL (n, node_exists_on_other_branch) {
    WITH n, node_exists_on_other_branch
    WHERE node_exists_on_other_branch = FALSE
    OPTIONAL MATCH (n)-[:IS_RELATED {branch: $global_branch_name}]-(rel:Relationship)
    DETACH DELETE rel
} IN TRANSACTIONS OF 500 ROWS
CALL (n, node_exists_on_other_branch) {
    WITH n, node_exists_on_other_branch
    WHERE node_exists_on_other_branch = FALSE
    OPTIONAL MATCH (n)-[:HAS_ATTRIBUTE {branch: $global_branch_name}]-(attr:Attribute)
    DETACH DELETE attr
} IN TRANSACTIONS OF 500 ROWS

// reduce the results to a single row
WITH 1 AS one
LIMIT 1

// --------------
// for every edge on this branch, delete it
// --------------
MATCH (s)-[r]->(d)
WHERE r.branch = $branch_name
CALL (r) {
    DELETE r
} IN TRANSACTIONS OF 500 ROWS

// --------------
// get the database IDs of every vertex linked to a deleted edge
// --------------
WITH DISTINCT elementId(s) AS s_id, elementId(d) AS d_id
WITH collect(s_id) + collect(d_id) AS vertex_ids
UNWIND vertex_ids AS vertex_id

// --------------
// delete any vertices that are now orphaned
// --------------
CALL (vertex_id) {
    MATCH (n)
    WHERE elementId(n) = vertex_id
    AND NOT exists((n)--())
    DELETE n
} IN TRANSACTIONS OF 500 ROWS
        """
        self.params["branch_name"] = self.branch_name
        self.params["global_branch_name"] = GLOBAL_BRANCH_NAME
        self.add_to_query(query)


class RebaseBranchQuery(Query):
    """Rebase a branch onto the default branch by updating edge timestamps

    For every edge on this branch
        if it has a from time before $at and no to time, update it to $at
        if it has a to time before $at, delete the edge
        if it has a to time after $at, update the from time to $at
    Then delete any orphaned vertices
    """

    name: str = "rebase_branch"
    type: QueryType = QueryType.WRITE
    insert_return: bool = False
    raise_error_if_empty: bool = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params["branch_name"] = self.branch.name
        self.params["at"] = self.at.to_string()

        query = """
// --------------
// Get all edges on this branch with their source and destination vertices
// --------------
MATCH (s)-[r]-(d)
WHERE r.branch = $branch_name
WITH DISTINCT r, s, d
WITH r, s, d,
    CASE
        // No `to` and `from` <= at: update
        WHEN r.to IS NULL AND r.from <= $at THEN TRUE
        // Has `to` and `to` < at: delete
        WHEN r.to IS NOT NULL AND r.to < $at THEN FALSE
        // Has `to` and `to` >= at: update
        ELSE TRUE
    END AS do_update

// --------------
// Process updates: set from = at for relationships we're keeping
// --------------
CALL (r, do_update) {
    WITH r, do_update
    WHERE do_update = TRUE
    SET r.from = $at
}

// --------------
// Delete the edges
// --------------
WITH r, s, d, do_update
WHERE do_update = FALSE
CALL (r, s, d) {
    DELETE r
}
// --------------
// Clean up any orpahned nodes edges
// --------------
WITH DISTINCT s, d
UNWIND [s, d] AS n
WITH DISTINCT n
CALL (n) {
    MATCH (n)
    WHERE NOT exists((n)--())
    DELETE n
}
        """
        self.add_to_query(query=query)


class BranchNodeGetListQuery(StandardNodeGetListQuery):
    def __init__(
        self,
        exclude_global: bool = False,
        branch_filters: BranchListFilters | None = None,
        **kwargs: Any,
    ) -> None:
        self.branch_filters = branch_filters or BranchListFilters()
        self.exclude_global = exclude_global

        # Temporary storage for filter params (will be merged after super().__init__)
        self._branch_filter_params: dict[str, Any] = {}

        # Build raw_filter from branch_filters
        self.raw_filter = self._build_raw_filter()

        # Pass name/ids/partial_match to parent for existing handling
        super().__init__(
            ids=self.branch_filters.ids,
            node_name=self.branch_filters.name,
            partial_match=self.branch_filters.partial_match,
            **kwargs,
        )

        # Merge our filter params into the query params
        self.params.update(self._branch_filter_params)

    def _build_raw_filter(self) -> str:
        """Build Cypher WHERE clause conditions from branch_filters."""
        conditions: list[str] = []

        # Always exclude DELETING branches
        conditions.append(f"n.status <> '{BranchStatus.DELETING.value}'")

        if self.exclude_global:
            conditions.append(f"n.name <> '{GLOBAL_BRANCH_NAME}'")

        if self.branch_filters.status:
            param_name = "filter_status"
            self._branch_filter_params[param_name] = self.branch_filters.status.value
            conditions.append(f"n.status = ${param_name}")

        if self.branch_filters.created_by_id:
            param_name = "filter_created_by"
            self._branch_filter_params[param_name] = self.branch_filters.created_by_id
            conditions.append(f"n.created_by = ${param_name}")

        # Branched from (rebase timestamp) filters (with NULL check)
        if self.branch_filters.branched_from_after:
            param_name = "filter_branched_from_after"
            self._branch_filter_params[param_name] = Timestamp(
                self.branch_filters.branched_from_after.isoformat()
            ).to_string()
            conditions.append(f"(n.branched_from IS NOT NULL AND n.branched_from > ${param_name})")

        if self.branch_filters.branched_from_before:
            param_name = "filter_branched_from_before"
            self._branch_filter_params[param_name] = Timestamp(
                self.branch_filters.branched_from_before.isoformat()
            ).to_string()
            conditions.append(f"(n.branched_from IS NOT NULL AND n.branched_from < ${param_name})")

        if self.branch_filters.created_at_after:
            param_name = "filter_created_at_after"
            self._branch_filter_params[param_name] = Timestamp(
                self.branch_filters.created_at_after.isoformat()
            ).to_string()
            conditions.append(f"(n.created_at IS NOT NULL AND n.created_at > ${param_name})")

        if self.branch_filters.created_at_before:
            param_name = "filter_created_at_before"
            self._branch_filter_params[param_name] = Timestamp(
                self.branch_filters.created_at_before.isoformat()
            ).to_string()
            conditions.append(f"(n.created_at IS NOT NULL AND n.created_at < ${param_name})")

        if self.branch_filters.updated_at_after:
            param_name = "filter_updated_at_after"
            self._branch_filter_params[param_name] = Timestamp(
                self.branch_filters.updated_at_after.isoformat()
            ).to_string()
            conditions.append(f"(n.updated_at IS NOT NULL AND n.updated_at > ${param_name})")

        if self.branch_filters.updated_at_before:
            param_name = "filter_updated_at_before"
            self._branch_filter_params[param_name] = Timestamp(
                self.branch_filters.updated_at_before.isoformat()
            ).to_string()
            conditions.append(f"(n.updated_at IS NOT NULL AND n.updated_at < ${param_name})")

        return " AND ".join(conditions) if conditions else ""



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Any

import ujson
from rich.progress import Progress, TaskID

from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME, BranchSupportType, SchemaPathType
from infrahub.core.initialization import get_root_node
from infrahub.core.migrations.graph.m044_backfill_hfid_display_label_in_db import (
    DefaultBranchNodeCount,
    GetPathDetailsDefaultBranch,
    GetResultMapQuery,
    UpdateAttributeValuesQuery,
)
from infrahub.core.migrations.schema.node_attribute_add import NodeAttributeAddMigration
from infrahub.core.migrations.shared import ArbitraryMigration, MigrationInput, MigrationResult, get_migration_console
from infrahub.core.path import SchemaPath
from infrahub.core.query import Query, QueryType

from .load_schema_branch import get_or_load_schema_branch

if TYPE_CHECKING:
    from infrahub.core.schema import AttributeSchema, MainSchemaTypes, NodeSchema, SchemaAttributePath
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase


class DeleteBranchAwareAttrsForBranchAgnosticNodesQuery(Query):
    name = "delete_branch_aware_attrs_for_branch_agnostic_nodes_query"
    type = QueryType.WRITE
    insert_return = False
    raise_error_if_empty = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (n:Node {branch_support: "agnostic"})
MATCH (n)-[:HAS_ATTRIBUTE]->(attr:Attribute)
WHERE attr.name IN ["human_friendly_id", "display_label"]
WITH DISTINCT attr
CALL (attr) {
    DETACH DELETE attr
} IN TRANSACTIONS
        """
        self.add_to_query(query)


class Migration046(ArbitraryMigration):
    """
    Delete any branch-aware human_friendly_id and display_label attributes added to branch-agnostic nodes
    Add human_friendly_id and display_label attributes to branch-agnostic nodes
    Set human_friendly_id and display_label attributes for branch-agnostic nodes on global branch

    Uses and duplicates code from Migration044
    """

    name: str = "046_fill_agnostic_hfid_display_labels"
    minimum_version: int = 45
    update_batch_size: int = 1000

    async def _do_one_schema_all(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        schema: MainSchemaTypes,
        schema_branch: SchemaBranch,
        attribute_schema_map: dict[AttributeSchema, AttributeSchema],
        progress: Progress | None = None,
        update_task: TaskID | None = None,
    ) -> None:
        print(f"Processing {schema.kind}...", end="")

        schema_paths_by_name: dict[str, list[SchemaAttributePath]] = {}
        for source_attribute_schema in attribute_schema_map.keys():
            node_schema_property = getattr(schema, source_attribute_schema.name)
            if not node_schema_property:
                continue
            if isinstance(node_schema_property, list):
                schema_paths_by_name[source_attribute_schema.name] = [
                    schema.parse_schema_path(path=str(path), schema=schema_branch) for path in node_schema_property
                ]
            else:
                schema_paths_by_name[source_attribute_schema.name] = [
                    schema.parse_schema_path(path=str(node_schema_property), schema=schema_branch)
                ]
        all_schema_paths = list(chain(*schema_paths_by_name.values()))
        offset = 0

        # loop until we get no results from the get_details_query
        while True:
            get_details_query: GetResultMapQuery = await GetPathDetailsDefaultBranch.init(
                db=db,
                schema_kind=schema.kind,
                schema_paths=all_schema_paths,
                offset=offset,
                limit=self.update_batch_size,
            )
            await get_details_query.execute(db=db)

            num_updates = 0
            for source_attribute_schema, destination_attribute_schema in attribute_schema_map.items():
                schema_paths = schema_paths_by_name[source_attribute_schema.name]
                schema_path_values_map = get_details_query.get_result_map(schema_paths)
                num_updates = max(num_updates, len(schema_path_values_map))
                formatted_schema_path_values_map = {}
                for k, v in schema_path_values_map.items():
                    if not v:
                        continue
                    if destination_attribute_schema.kind == "List":
                        formatted_schema_path_values_map[k] = ujson.dumps(v)
                    else:
                        formatted_schema_path_values_map[k] = " ".join(item for item in v if item is not None)

                if not formatted_schema_path_values_map:
                    continue

                update_display_label_query = await UpdateAttributeValuesQuery.init(
                    db=db,
                    branch=branch,
                    attribute_schema=destination_attribute_schema,
                    values_by_id_map=formatted_schema_path_values_map,
                )
                await update_display_label_query.execute(db=db)

            if progress is not None and update_task is not None:
                progress.update(update_task, advance=num_updates)

            if num_updates == 0:
                break

            offset += self.update_batch_size

        print("done")

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        try:
            return await self._do_execute(migration_input=migration_input)
        except Exception as exc:
            return MigrationResult(errors=[str(exc)])

    async def _do_execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        console = get_migration_console()
        result = MigrationResult()

        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)
        main_schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)

        console.print("Deleting branch-aware attributes for branch-agnostic nodes...", end="")
        delete_query = await DeleteBranchAwareAttrsForBranchAgnosticNodesQuery.init(db=db)
        await delete_query.execute(db=db)
        console.print("done")

        branch_agnostic_schemas: list[NodeSchema] = []
        migrations = []
        for node_schema_kind in main_schema_branch.node_names:
            schema = main_schema_branch.get_node(name=node_schema_kind, duplicate=False)
            if schema.branch is not BranchSupportType.AGNOSTIC:
                continue
            branch_agnostic_schemas.append(schema)
            migrations.extend(
                [
                    NodeAttributeAddMigration(
                        new_node_schema=schema,
                        previous_node_schema=schema,
                        schema_path=SchemaPath(
                            schema_kind=schema.kind, path_type=SchemaPathType.ATTRIBUTE, field_name="human_friendly_id"
                        ),
                    ),
                    NodeAttributeAddMigration(
                        new_node_schema=schema,
                        previous_node_schema=schema,
                        schema_path=SchemaPath(
                            schema_kind=schema.kind, path_type=SchemaPathType.ATTRIBUTE, field_name="display_label"
                        ),
                    ),
                ]
            )

        global_branch = await Branch.get_by_name(db=db, name=GLOBAL_BRANCH_NAME)
        with Progress(console=console) as progress:
            update_task = progress.add_task(
                "Adding HFID and display label attributes to branch-agnostic nodes", total=len(migrations)
            )

            for migration in migrations:
                try:
                    execution_result = await migration.execute(migration_input=migration_input, branch=global_branch)
                    result.errors.extend(execution_result.errors)
                    progress.update(update_task, advance=1)
                except Exception as exc:
                    result.errors.append(str(exc))
                    return result

        total_nodes_query = await DefaultBranchNodeCount.init(
            db=db, kinds_to_include=[sch.kind for sch in branch_agnostic_schemas]
        )
        await total_nodes_query.execute(db=db)
        total_nodes_count = total_nodes_query.get_num_nodes()

        base_node_schema = main_schema_branch.get("SchemaNode", duplicate=False)
        display_label_attribute_schema = base_node_schema.get_attribute("display_label")
        display_labels_attribute_schema = base_node_schema.get_attribute("display_labels")
        hfid_attribute_schema = base_node_schema.get_attribute("human_friendly_id")

        with Progress(console=console) as progress:
            update_task = progress.add_task(
                f"Set display_label and human_friendly_id for {total_nodes_count} branch-agnostic nodes on global branch",
                total=total_nodes_count,
            )
            for branch_agnostic_schema in branch_agnostic_schemas:
                attribute_schema_map = {}
                if branch_agnostic_schema.display_labels:
                    attribute_schema_map[display_labels_attribute_schema] = display_label_attribute_schema
                if branch_agnostic_schema.human_friendly_id:
                    attribute_schema_map[hfid_attribute_schema] = hfid_attribute_schema
                if not attribute_schema_map:
                    continue

                await self._do_one_schema_all(
                    db=db,
                    branch=global_branch,
                    schema=branch_agnostic_schema,
                    schema_branch=main_schema_branch,
                    attribute_schema_map=attribute_schema_map,
                    progress=progress,
                    update_task=update_task,
                )

        return result

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.constants import RelationshipStatus
from infrahub.core.graph.schema import GraphNodeRelationships, GraphRelDirection

from ..query import MigrationQuery
from ..shared import SchemaMigration

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo

    from infrahub.database import InfrahubDatabase


class NodeRemoveMigrationBaseQuery(MigrationQuery):
    def render_sub_query_per_rel_type(
        self,
        rel_name: str,
        rel_type: str,
        rel_def: FieldInfo,
    ) -> str:
        subquery = [
            f"WITH peer_node, {rel_name}, active_node",
            f'WHERE type({rel_name}) = "{rel_type}"',
        ]
        if rel_def.default.direction in [GraphRelDirection.OUTBOUND, GraphRelDirection.EITHER]:
            subquery.append(f"""
                CREATE (active_node)-[edge:{rel_type} $rel_props ]->(peer_node)
                SET edge.branch = CASE WHEN {rel_name}.branch = "-global-" THEN "-global-" ELSE $branch END
                SET edge.branch_level = CASE WHEN {rel_name}.branch = "-global-" THEN {rel_name}.branch_level ELSE $branch_level END
                """)
        elif rel_def.default.direction in [GraphRelDirection.INBOUND, GraphRelDirection.EITHER]:
            subquery.append(f"""
                CREATE (active_node)<-[edge:{rel_type} $rel_props ]-(peer_node)
                SET edge.branch = CASE WHEN {rel_name}.branch = "-global-" THEN "-global-" ELSE $branch END
                SET edge.branch_level = CASE WHEN {rel_name}.branch = "-global-" THEN {rel_name}.branch_level ELSE $branch_level END
                """)
        subquery.append("RETURN peer_node as p2")
        return "\n".join(subquery)

    def render_node_remove_query(self, branch_filter: str) -> str:
        raise NotImplementedError()

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["current_time"] = self.at.to_string()
        self.params["branch_name"] = self.branch.name
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["user_id"] = self.user_id

        self.params["rel_props"] = {
            "status": RelationshipStatus.DELETED.value,
            "from": self.at.to_string(),
            "from_user_id": self.user_id,
        }

        # Set metadata for vertex properties on default/global branch
        self.params["set_metadata"] = self.branch.is_default or self.branch.is_global

        node_remove_query = self.render_node_remove_query(branch_filter=branch_filter)

        query = """
        // Find all the active nodes
        MATCH (node:%(node_kind)s)
        CALL (node) {
            MATCH (root:Root)<-[r:IS_PART_OF]-(node)
            WHERE %(branch_filter)s
            RETURN node as n1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, r1 as rb
        WHERE rb.status = "active"
        %(node_remove_query)s
        WITH active_node
        // Set metadata on Node vertex if on default/global branch
        CALL (active_node) {
            WITH active_node
            WHERE $set_metadata
            SET active_node.updated_at = $current_time, active_node.updated_by = $user_id
        }
        RETURN DISTINCT active_node
        """ % {
            "branch_filter": branch_filter,
            "node_remove_query": node_remove_query,
            "node_kind": self.migration.previous_schema.kind,
        }
        self.add_to_query(query)

    def get_nbr_migrations_executed(self) -> int:
        return self.stats.get_counter(name="nodes_created")


class NodeRemoveMigrationQueryIn(NodeRemoveMigrationBaseQuery):
    name = "migration_node_remove_in"
    insert_return: bool = False

    def render_node_remove_query(self, branch_filter: str) -> str:
        sub_query, sub_query_args = self.render_sub_query_in()
        query = """
        // Process Inbound Relationship
        WITH active_node
        MATCH (active_node)<-[]-(peer)
        CALL (active_node, peer) {
            MATCH (active_node)-[r]->(peer)
            WHERE %(branch_filter)s
            RETURN active_node as n1, r as rel_inband1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, rel_inband1 as rel_inband, p1 as peer_node
        WHERE rel_inband.status = "active"
        CALL (peer_node) {
            WITH peer_node
            WHERE $set_metadata
            SET peer_node.updated_at = $current_time, peer_node.updated_by = $user_id
        }
        CALL (%(sub_query_args)s) {
            %(sub_query)s
        }
        WITH p2 as peer_node, rel_inband, active_node
        FOREACH (i in CASE WHEN rel_inband.branch IN ["-global-", $branch] THEN [1] ELSE [] END |
            SET rel_inband.to = $current_time, rel_inband.to_user_id = $user_id
        )
        """ % {"sub_query": sub_query, "sub_query_args": sub_query_args, "branch_filter": branch_filter}
        return query

    def render_sub_query_in(self) -> tuple[str, str]:
        rel_name = "rel_inband"
        sub_query_in_args = f"peer_node, {rel_name}, active_node"
        sub_queries_in = [
            self.render_sub_query_per_rel_type(
                rel_name=rel_name,
                rel_type=rel_type,
                rel_def=rel_def,
            )
            for rel_type, rel_def in GraphNodeRelationships.model_fields.items()
        ]
        sub_query_in = "\nUNION\n".join(sub_queries_in)
        return sub_query_in, sub_query_in_args

    def get_nbr_migrations_executed(self) -> int:
        return 0


class NodeRemoveMigrationQueryOut(NodeRemoveMigrationBaseQuery):
    name = "migration_node_remove_in"
    insert_return: bool = False

    def render_node_remove_query(self, branch_filter: str) -> str:
        sub_query, sub_query_args = self.render_sub_query_out()
        query = """
        // Process Outbound Relationship
        WITH active_node
        MATCH (active_node)-[]->(peer)
        CALL (active_node, peer) {
            MATCH (active_node)-[r]->(peer)
            WHERE %(branch_filter)s
            RETURN active_node as n1, r as rel_outband1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, rel_outband1 as rel_outband, p1 as peer_node
        WHERE rel_outband.status = "active"
        CALL (peer_node) {
            WITH peer_node
            WHERE $set_metadata
            SET peer_node.updated_at = $current_time, peer_node.updated_by = $user_id
        }
        CALL (%(sub_query_args)s) {
            %(sub_query)s
        }
        FOREACH (i in CASE WHEN rel_outband.branch IN ["-global-", $branch] THEN [1] ELSE [] END |
            SET rel_outband.to = $current_time, rel_outband.to_user_id = $user_id
        )
        """ % {"sub_query": sub_query, "sub_query_args": sub_query_args, "branch_filter": branch_filter}

        return query

    def render_sub_query_out(self) -> tuple[str, str]:
        rel_name = "rel_outband"
        sub_query_out_args = f"peer_node, {rel_name}, active_node"
        sub_queries_out = [
            self.render_sub_query_per_rel_type(
                rel_name=rel_name,
                rel_type=rel_type,
                rel_def=rel_def,
            )
            for rel_type, rel_def in GraphNodeRelationships.model_fields.items()
        ]
        sub_query_out = "\nUNION\n".join(sub_queries_out)
        return sub_query_out, sub_query_out_args

    def get_nbr_migrations_executed(self) -> int:
        return self.num_of_results


class NodeRemoveMigration(SchemaMigration):
    name: str = "node.remove"
    queries: Sequence[type[MigrationQuery]] = [NodeRemoveMigrationQueryIn, NodeRemoveMigrationQueryOut]  # type: ignore[assignment]



from __future__ import annotations

import asyncio
from collections import defaultdict
from typing import TYPE_CHECKING

from typing_extensions import Self

from infrahub.core.constants import (
    DiffAction,
    InfrahubKind,
)
from infrahub.core.manager import NodeManager
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import DiffFromRequiredOnDefaultBranchError, DiffRangeValidationError

from ...git.models import GitDiffNamesOnly
from ...workflows.catalogue import GIT_REPOSITORIES_DIFF_NAMES_ONLY
from .model.diff import (
    FileDiffElement,
)

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase
    from infrahub.services import InfrahubServices

    from ..branch import Branch
    from ..node import Node


class BranchDiffer:
    diff_from: Timestamp
    diff_to: Timestamp

    def __init__(
        self,
        branch: Branch,
        origin_branch: Branch | None = None,
        branch_only: bool = False,
        diff_from: str | Timestamp | None = None,
        diff_to: str | Timestamp | None = None,
        db: InfrahubDatabase | None = None,
        service: InfrahubServices | None = None,
    ) -> None:
        """_summary_

        Args:
            branch (Branch): Main branch this diff is caculated from
            origin_branch (Branch): Storing the origin branch the main branch started from for convenience.
            branch_only (bool, optional): When True, only consider the changes in the branch, ignore the changes in main. Defaults to False.
            diff_from (Union[str, Timestamp], optional): Time from when the diff is calculated. Defaults to None.
            diff_to (Union[str, Timestamp], optional): Time to when the diff is calculated. Defaults to None.

        Raises:
            ValueError: if diff_from and diff_to are not correct
        """

        self.branch = branch
        self.branch_only = branch_only
        self.origin_branch = origin_branch

        self._db = db
        self._service = service

        if not diff_from and self.branch.is_default:
            raise DiffFromRequiredOnDefaultBranchError(
                f"diff_from is mandatory when diffing on the default branch `{self.branch.name}`."
            )

        # If diff from hasn't been provided, we'll use the creation of the branch as the starting point
        if diff_from:
            self.diff_from = Timestamp(diff_from)
        else:
            self.diff_from = Timestamp(self.branch.created_at)

        # If diff_to hasn't been provided, we will use the current time.
        self.diff_to = Timestamp(diff_to)

        if self.diff_to < self.diff_from:
            raise DiffRangeValidationError("diff_to must be later than diff_from")

        # Results organized by Branch
        self._results: dict[str, dict] = defaultdict(lambda: {"nodes": {}, "rels": defaultdict(dict), "files": {}})
        self._calculated_diff_files_at: Timestamp | None = None

    @property
    def service(self) -> InfrahubServices:
        if not self._service:
            raise ValueError("BranchDiffer object was not initialized with InfrahubServices")
        return self._service

    @property
    def db(self) -> InfrahubDatabase:
        if not self._db:
            raise ValueError("BranchDiffer object was not initialized with InfrahubDatabase")
        return self._db

    @classmethod
    async def init(
        cls,
        db: InfrahubDatabase,
        branch: Branch,
        branch_only: bool = False,
        diff_from: str | Timestamp | None = None,
        diff_to: str | Timestamp | None = None,
        service: InfrahubServices | None = None,
    ) -> Self:
        origin_branch = branch.get_origin_branch()

        return cls(
            branch=branch,
            origin_branch=origin_branch,
            branch_only=branch_only,
            diff_from=diff_from,
            diff_to=diff_to,
            db=db,
            service=service,
        )

    async def get_files(self) -> dict[str, list[FileDiffElement]]:
        if not self._calculated_diff_files_at:
            await self._calculated_diff_files()

        return {
            branch_name: data["files"]
            for branch_name, data in self._results.items()
            if not self.branch_only or branch_name == self.branch.name
        }

    async def _calculated_diff_files(self) -> None:
        self._results[self.branch.name]["files"] = await self.get_files_repositories_for_branch(branch=self.branch)

        if self.origin_branch:
            self._results[self.origin_branch.name]["files"] = await self.get_files_repositories_for_branch(
                branch=self.origin_branch
            )

        self._calculated_diff_files_at = Timestamp()

    async def get_files_repository(
        self,
        branch_name: str,
        repository: Node,
        commit_from: str,
        commit_to: str,
    ) -> list[FileDiffElement]:
        """Return all the files that have added, changed or removed for a given repository between 2 commits."""

        files = []

        model = GitDiffNamesOnly(
            repository_id=repository.id,
            repository_name=repository.name.value,  # type: ignore[attr-defined]
            repository_kind=repository.get_kind(),
            first_commit=commit_from,
            second_commit=commit_to,
        )

        diff = await self.service.workflow.execute_workflow(
            workflow=GIT_REPOSITORIES_DIFF_NAMES_ONLY, parameters={"model": model}
        )

        actions = {
            "files_changed": DiffAction.UPDATED,
            "files_added": DiffAction.ADDED,
            "files_removed": DiffAction.REMOVED,
        }

        for action_name, diff_action in actions.items():
            for filename in getattr(diff, action_name, []):
                files.append(
                    FileDiffElement(
                        branch=branch_name,
                        location=filename,
                        repository=repository,
                        action=diff_action,
                        commit_to=commit_to,
                        commit_from=commit_from,
                    )
                )

        return files

    async def get_files_repositories_for_branch(self, branch: Branch) -> list[FileDiffElement]:
        tasks = []
        files = []

        repos_to = {
            repo.id: repo
            for repo in await NodeManager.query(
                schema=InfrahubKind.GENERICREPOSITORY, db=self.db, branch=branch, at=self.diff_to
            )
        }
        repos_from = {
            repo.id: repo
            for repo in await NodeManager.query(
                schema=InfrahubKind.GENERICREPOSITORY, db=self.db, branch=branch, at=self.diff_from
            )
        }

        # For now we are ignoring the repos that are either not present at to time or at from time.
        # These repos will be identified in the graph already
        repo_ids_common = set(repos_to.keys()) & set(repos_from.keys())

        for repo_id in repo_ids_common:
            if repos_to[repo_id].commit.value == repos_from[repo_id].commit.value:  # type: ignore[attr-defined]
                continue

            tasks.append(
                self.get_files_repository(
                    branch_name=branch.name,
                    repository=repos_to[repo_id],
                    commit_from=repos_from[repo_id].commit.value,  # type: ignore[attr-defined]
                    commit_to=repos_to[repo_id].commit.value,  # type: ignore[attr-defined]
                )
            )

        responses = await asyncio.gather(*tasks)

        for response in responses:
            if isinstance(response, list):
                files.extend(response)

        return files



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.constants import NULL_VALUE
from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class Migration055Query01(Query):
    """Remove the default_value from CoreWebhook's validate_certificates attribute.

    This migration finds the CoreWebhook SchemaGeneric, locates its validate_certificates
    SchemaAttribute, and sets the default_value to NULL. This removes the previous default
    of True.
    """

    name = "migration_055_01"
    type: QueryType = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["at"] = self.at.to_string()
        self.params["null_value"] = NULL_VALUE

        query = """
// get the generic schema
MATCH p1 = (sg:SchemaGeneric)-[:HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[:HAS_VALUE]->(:AttributeValueIndexed {value: "Webhook"})
WHERE all(r IN relationships(p1) WHERE r.status = "active" AND r.to IS NULL)

// for safety, also check that the sg is in the namespace "Core"
MATCH p2 = (sg)-[:HAS_ATTRIBUTE]->(:Attribute {name: "namespace"})-[:HAS_VALUE]->(:AttributeValueIndexed {value: "Core"})
WHERE all(r IN relationships(p2) WHERE r.status = "active" AND r.to IS NULL)

// there should only be 1 CoreWebhook schema generic
WITH sg
LIMIT 1

// find the validate_certificates attribute
MATCH p3 = (sg)-[:IS_RELATED]-(:Relationship {name: "schema__node__attributes"})
    -[:IS_RELATED]-(sa:SchemaAttribute)
    -[:HAS_ATTRIBUTE]->(:Attribute {name: "name"})
    -[:HAS_VALUE]->(:AttributeValueIndexed {value: "validate_certificates"})
WHERE all(r IN relationships(p3) WHERE r.status = "active" AND r.to IS NULL)
// there should only be 1 validate_certificates attribute
WITH sa
LIMIT 1

// get the default_value Attribute
MATCH (sa)-[ha:HAS_ATTRIBUTE]->(default_value_attr:Attribute {name: "default_value"})-[hv:HAS_VALUE]->(default_value)
WHERE all(r IN [ha, hv] WHERE r.status = "active" AND r.to IS NULL)
LIMIT 1

// skip if it is already NULL
WITH sa, default_value_attr, hv, default_value
WHERE default_value.value <> $null_value

// close the HAS_VALUE edge for the current default value
SET hv.to = $at

// get the new value
MERGE (new_value:AttributeValue:AttributeValueIndexed {value: $null_value, is_default: true})
LIMIT 1

// link the new value
CREATE (default_value_attr)-[new_hv:HAS_VALUE]->(new_value)
SET new_hv = properties(hv)
SET new_hv.from = $at, new_hv.to = NULL
        """
        self.add_to_query(query)
        self.return_labels = ["new_value"]


class Migration055(GraphMigration):
    name: str = "055_remove_webhook_validate_certificates_default"
    queries: Sequence[type[Query]] = [Migration055Query01]
    minimum_version: int = 54

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()

        return result



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.constants import (
    OBJECT_TEMPLATE_RELATIONSHIP_NAME,
    AllowOverrideType,
    InfrahubKind,
    RelationshipKind,
)

from .generated.node_schema import GeneratedNodeSchema
from .generic_schema import GenericSchema

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase


class NodeSchema(GeneratedNodeSchema):
    @property
    def is_node_schema(self) -> bool:
        return True

    @property
    def is_generic_schema(self) -> bool:
        return False

    @property
    def is_profile_schema(self) -> bool:
        return False

    @property
    def is_template_schema(self) -> bool:
        return False

    @property
    def is_ip_prefix(self) -> bool:
        """Return whether a node is a derivative of built-in IP prefixes."""
        return InfrahubKind.IPPREFIX in self.inherit_from

    @property
    def is_ip_address(self) -> bool:
        """Return whether a node is a derivative of built-in IP addreses."""
        return InfrahubKind.IPADDRESS in self.inherit_from

    @property
    def is_file_object(self) -> bool:
        """Return whether a node is a derivative of built-in file objects."""
        return InfrahubKind.FILEOBJECT in self.inherit_from

    def validate_inheritance(self, interface: GenericSchema) -> None:
        """Perform checks specific to inheritance from Generics.

        Checks:
            - Check that protected attributes and relationships are not overridden before inheriting them from interface.
            - Check that the attribute types to be inherited are same kind.
        """
        for attribute in self.attributes:
            if attribute.name in interface.attribute_names:
                if (
                    not attribute.inherited
                    and interface.get_attribute(attribute.name).allow_override == AllowOverrideType.NONE
                ):
                    raise ValueError(
                        f"{self.kind}'s attribute {attribute.name} inherited from {interface.kind} cannot be overriden"
                    )

                interface_attr = interface.get_attribute(attribute.name)
                # Check existing inherited attribute kind is the same as the incoming inherited attribute
                if attribute.kind != interface_attr.kind:
                    raise ValueError(
                        f"{self.kind}.{attribute.name} inherited from {interface.namespace}{interface.name} must be the same kind "
                        f'["{interface_attr.kind}", "{attribute.kind}"]'
                    )

        for relationship in self.relationships:
            if relationship.name in interface.relationship_names and not relationship.inherited:
                interface_relationship = interface.get_relationship(relationship.name)
                if interface_relationship.allow_override == AllowOverrideType.NONE:
                    raise ValueError(
                        f"{self.kind}'s relationship {relationship.name} inherited from {interface.kind} cannot be overriden"
                    )
                if (
                    relationship.kind == RelationshipKind.HIERARCHY
                    or relationship.name == OBJECT_TEMPLATE_RELATIONSHIP_NAME
                ):
                    continue
                if relationship.peer != interface_relationship.peer:
                    raise ValueError(
                        f"{self.kind}'s relationship {relationship.name} inherited from {interface.kind} must have the same peer "
                        f"({interface_relationship.peer} != {relationship.peer})"
                    )

    def get_hierarchy_schema(
        self, db: InfrahubDatabase, branch: Branch | str | None = None, duplicate: bool = False
    ) -> GenericSchema:
        if not self.hierarchy:
            raise ValueError("The node is not part of a hierarchy")
        schema = db.schema.get(name=self.hierarchy, branch=branch, duplicate=duplicate)
        if not isinstance(schema, GenericSchema):
            raise TypeError
        return schema

    def get_labels(self) -> list[str]:
        """Return the labels for this object, composed of the kind
        and the list of Generic this object is inheriting from."""

        labels: list[str] = [self.kind] + self.inherit_from
        if self.namespace not in ["Schema", "Internal"] and InfrahubKind.GENERICGROUP not in self.inherit_from:
            labels.append(InfrahubKind.NODE)
        return labels



from __future__ import annotations

from enum import Enum
from typing import Self

from infrahub.core.constants import InfrahubKind
from infrahub.core.schema.dropdown import DropdownChoice
from infrahub.utils import InfrahubStringEnum


class NodeAction(InfrahubStringEnum):
    CREATED = "created"
    UPDATED = "updated"


class DropdownEnum(Enum):
    @classmethod
    def available_types(cls) -> list[DropdownChoice]:
        return [cls.__members__[member].value for member in list(cls.__members__)]

    @classmethod
    def from_value(cls, value: str) -> Self:
        for member in cls.__members__:
            if value == cls.__members__[member].value.name:
                return cls.__members__[member]

        raise NotImplementedError(f"The defined value {value} doesn't match a value of {cls.__class__.__name__}")


class BranchScope(DropdownEnum):
    ALL_BRANCHES = DropdownChoice(
        name="all_branches",
        label="All Branches",
        description="All branches",
        color="#4cd964",
    )
    DEFAULT_BRANCH = DropdownChoice(
        name="default_branch",
        label="Default Branch",
        description="Only the default branch",
        color="#5ac8fa",
    )
    OTHER_BRANCHES = DropdownChoice(
        name="other_branches",
        label="Other Branches",
        description="All branches except the default branch",
        color="#ff2d55",
    )


class MemberAction(DropdownEnum):
    ADD_MEMBER = DropdownChoice(
        name="add_member",
        label="Add member",
        description="Add impacted member to the selected group",
        color="#4cd964",
    )
    REMOVE_MEMBER = DropdownChoice(
        name="remove_member",
        label="Remove member",
        description="Remove impacted member from the selected group",
        color="#ff2d55",
    )


class MemberUpdate(DropdownEnum):
    ADDED = DropdownChoice(
        name="added",
        label="Added",
        description="Trigger when members are added to this group",
        color="#4cd964",
    )
    REMOVED = DropdownChoice(
        name="removed",
        label="Removed",
        description="Trigger when members are removed from this group",
        color="#ff2d55",
    )


class RelationshipMatch(DropdownEnum):
    ADDED = DropdownChoice(
        name="added",
        label="Added",
        description="Check if the selected relationship was added",
        color="#4cd964",
    )
    REMOVED = DropdownChoice(
        name="removed",
        label="Removed",
        description="Check if the selected relationship was removed",
        color="#ff2d55",
    )
    UPDATED = DropdownChoice(
        name="updated",
        label="Updated",
        description="Check if the selected relationship was updated, added or removed.",
        color="#5ac8fa",
    )


class ValueMatch(DropdownEnum):
    VALUE = DropdownChoice(
        name="value",
        label="Value",
        description="Match against the current value",
        color="#4cd964",
    )
    VALUE_PREVIOUS = DropdownChoice(
        name="value_previous",
        label="Value Previous",
        description="Match against the previous value",
        color="#ff2d55",
    )
    VALUE_FULL = DropdownChoice(
        name="value_full",
        label="Full value match",
        description="Match against both the current and previous values",
        color="#5ac8fa",
    )
    ANY = DropdownChoice(
        name="any",
        label="Any",
        description="Match against any value",
        color="#276cc2",
    )


NODES_THAT_TRIGGER_ACTION_RULES_SETUP = [
    InfrahubKind.GENERATORACTION,
    InfrahubKind.GROUPACTION,
    InfrahubKind.GROUPTRIGGERRULE,
    InfrahubKind.NODETRIGGERRULE,
    InfrahubKind.NODETRIGGERATTRIBUTEMATCH,
    InfrahubKind.NODETRIGGERRELATIONSHIPMATCH,
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import datetime

    from infrahub.core.branch.enums import BranchStatus


@dataclass(slots=True)
class BranchListFilters:
    """Encapsulates all branch list query filters.

    This dataclass is constructed in the GraphQL layer and passed through
    to the core query layer. It follows the pattern established by
    StandardNodeOrdering for passing structured data through layers.

    Timestamp fields use datetime objects (converted from GraphQL DateTime scalar).
    The query layer converts these to Infrahub timestamp strings for Cypher comparison.
    """

    name: str | None = None
    ids: list[str] | None = field(default=None)
    partial_match: bool = False
    status: BranchStatus | None = None
    created_by_id: str | None = None
    branched_from_after: datetime | None = None
    branched_from_before: datetime | None = None
    created_at_after: datetime | None = None
    created_at_before: datetime | None = None
    updated_at_after: datetime | None = None
    updated_at_before: datetime | None = None



from typing import TYPE_CHECKING, Any

from graphql import GraphQLResolveInfo
from graphql.type.definition import GraphQLNonNull

from infrahub.core.branch.models import Branch
from infrahub.core.constants import BranchSupportType, MetadataOptions
from infrahub.core.manager import NodeManager
from infrahub.core.metadata.model import MetadataQueryOptions
from infrahub.core.node import Node
from infrahub.core.relationship import Relationship
from infrahub.core.schema.relationship_schema import RelationshipSchema
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase
from infrahub.graphql.field_extractor import extract_graphql_fields
from infrahub.graphql.metadata import build_metadata_query_options, get_metadata_options_from_fields

from ..loaders.node import GetManyParams, NodeDataLoader
from ..types import RELATIONS_PROPERTY_MAP, RELATIONS_PROPERTY_MAP_REVERSED

if TYPE_CHECKING:
    from infrahub.core.schema.node_schema import NodeSchema

    from ..initialization import GraphqlContext


class SingleRelationshipResolver:
    def __init__(self) -> None:
        self._data_loader_instances: dict[GetManyParams, NodeDataLoader] = {}

    def _build_relationship_meta_response(
        self, relationship: Relationship, metadata_fields: dict[str, Any]
    ) -> dict[str, Any]:
        data: dict[str, Any] = {}
        for meta_field in metadata_fields.keys():
            if meta_field == "created_at":
                created_at = relationship._get_created_at()
                data["created_at"] = created_at.to_datetime() if created_at else None
            elif meta_field == "created_by":
                account_id = relationship._get_created_by()
                data["created_by"] = {"id": account_id} if account_id else None
            elif meta_field == "updated_at":
                updated_at = relationship._get_updated_at()
                data["updated_at"] = updated_at.to_datetime() if updated_at else None
            elif meta_field == "updated_by":
                account_id = relationship._get_updated_by()
                data["updated_by"] = {"id": account_id} if account_id else None
        return data

    async def resolve(self, parent: dict, info: GraphQLResolveInfo, **kwargs: Any) -> dict[str, Any]:
        """Resolver for relationships of cardinality=one for Edged responses

        This resolver is used for paginated responses and as such we redefined the requested
        fields by only reusing information below the 'node' key.
        """
        # Extract the InfraHub schema by inspecting the GQL Schema

        # :
        node_schema: NodeSchema = (
            info.parent_type.of_type.graphene_type._meta.schema
            if isinstance(info.parent_type, GraphQLNonNull)
            else info.parent_type.graphene_type._meta.schema  # type: ignore[attr-defined]
        )

        graphql_context: GraphqlContext = info.context

        # Extract the name of the fields in the GQL query
        fields = extract_graphql_fields(info=info)
        node_fields = fields.get("node", {})
        property_fields = fields.get("properties", {})
        metadata_fields = {
            "node_metadata": fields.get("node_metadata", {}),
            "relationship_metadata": fields.get("relationship_metadata", {}),
        }
        for key, value in property_fields.items():
            mapped_name = RELATIONS_PROPERTY_MAP[key]
            node_fields[mapped_name] = value

        metadata_field_names = {prop_name for prop_name in RELATIONS_PROPERTY_MAP if prop_name != "__typename"}
        requires_relationship_properties = bool(set(property_fields.keys()) & metadata_field_names)
        requires_relationship_metadata = bool(metadata_fields["relationship_metadata"])

        # Extract the schema of the node on the other end of the relationship from the GQL Schema
        node_rel = node_schema.get_relationship(info.field_name)

        response: dict[str, Any] = {"node": None, "properties": {}}

        relationship: Relationship | None = None
        peer_node: Node | None = None

        if requires_relationship_properties or requires_relationship_metadata:
            include_metadata = build_metadata_query_options(
                node_metadata_fields=metadata_fields.get("node_metadata"),
                relationship_metadata_fields=metadata_fields.get("relationship_metadata"),
                node_fields=node_fields,
            )
            # Add relationship properties metadata to relationship_level
            include_metadata |= MetadataQueryOptions(
                relationship_level=get_metadata_options_from_fields(property_fields)
            )
            relationship = await self._get_entities_simple(
                db=graphql_context.db,
                branch=graphql_context.branch,
                at=graphql_context.at,
                field_name=info.field_name,
                parent_id=parent["id"],
                source_kind=node_schema.kind,
                rel_schema=node_rel,
                node_fields=node_fields,
                include_metadata=include_metadata,
                **kwargs,
            )
        else:
            include_metadata = build_metadata_query_options(
                node_metadata_fields=metadata_fields.get("node_metadata"),
                node_fields=node_fields,
            )
            peer_node = await self._get_entities_with_data_loader(
                db=graphql_context.db,
                branch=graphql_context.branch,
                at=graphql_context.at,
                rel_schema=node_rel,
                parent=parent,
                node_fields=node_fields,
                include_metadata=include_metadata,
            )

        if not relationship and not peer_node:
            return response

        async with graphql_context.db.start_session(read_only=True) as db:
            if relationship:
                node_graph = await relationship.to_graphql(
                    db=db, fields=node_fields, related_node_ids=graphql_context.related_node_ids
                )
                peer_node = await relationship.get_peer(db=db)
            elif peer_node:
                node_graph = await peer_node.to_graphql(
                    db=db, fields=node_fields, related_node_ids=graphql_context.related_node_ids
                )

            response["node"] = node_graph

            for key, mapped in RELATIONS_PROPERTY_MAP_REVERSED.items():
                value = node_graph.pop(key, None)
                if value:
                    response["properties"][mapped] = value

            if metadata_fields.get("node_metadata") and peer_node:
                response["node_metadata"] = await peer_node._build_meta_response("node_metadata", fields)

            if metadata_fields.get("relationship_metadata") and relationship:
                response["relationship_metadata"] = self._build_relationship_meta_response(
                    relationship=relationship, metadata_fields=metadata_fields["relationship_metadata"]
                )

        return response

    async def _get_entities_simple(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp | None,
        field_name: str,
        parent_id: str,
        source_kind: str,
        rel_schema: RelationshipSchema,
        node_fields: dict[str, Any],
        include_metadata: MetadataQueryOptions,
        **kwargs: Any,
    ) -> Relationship | None:
        filters = {
            f"{field_name}__{key}": value
            for key, value in kwargs.items()
            if ("__" in key and value) or key in ["id", "ids"]
        }
        async with db.start_session(read_only=True) as dbs:
            objs = await NodeManager.query_peers(
                db=dbs,
                ids=[parent_id],
                source_kind=source_kind,
                schema=rel_schema,
                filters=filters,
                fields=node_fields,
                at=at,
                branch=branch,
                branch_agnostic=rel_schema.branch is BranchSupportType.AGNOSTIC,
                fetch_peers=True,
                include_metadata=include_metadata,
            )
            if not objs:
                return None
            return objs[0]

    async def _get_entities_with_data_loader(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp | None,
        rel_schema: RelationshipSchema,
        parent: dict[str, Any],
        node_fields: dict[str, Any],
        include_metadata: MetadataQueryOptions,
    ) -> Node | None:
        try:
            peer_id: str = parent[rel_schema.name][0]["node"]["id"]
        except (KeyError, IndexError):
            return None

        if node_fields and "hfid" in node_fields:
            node_fields["human_friendly_id"] = None

        include_metadata |= MetadataQueryOptions(attribute_level=MetadataOptions.LINKED_NODES)
        query_params = GetManyParams(
            fields=node_fields,
            at=at,
            branch=branch,
            include_metadata=include_metadata,
            prefetch_relationships=False,
            branch_agnostic=rel_schema.branch is BranchSupportType.AGNOSTIC,
        )
        if query_params in self._data_loader_instances:
            loader = self._data_loader_instances[query_params]
        else:
            loader = NodeDataLoader(db=db, query_params=query_params)
            self._data_loader_instances[query_params] = loader
        node = await loader.load(key=peer_id)
        if not node:
            return None
        return node



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from infrahub.core.constants import BranchSupportType, RelationshipStatus
from infrahub.core.graph.schema import GraphAttributeRelationships
from infrahub.core.query import Query

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo

    from infrahub.database import InfrahubDatabase


class AttributeInfo(BaseModel):
    name: str
    node_kind: str
    branch_support: str = BranchSupportType.AWARE.value


class AttributeRenameQuery(Query):
    name = "attribute_rename"
    insert_return: bool = False

    def __init__(
        self,
        previous_attr: AttributeInfo,
        new_attr: AttributeInfo,
        **kwargs: Any,
    ) -> None:
        self.previous_attr = previous_attr
        self.new_attr = new_attr
        super().__init__(**kwargs)

    def render_match(self) -> str:
        query = """
        // Find all the active nodes
        CALL () {
            MATCH (node:%(node_kind)s)
            WHERE exists((node)-[:HAS_ATTRIBUTE]-(:Attribute { name: $prev_attr.name }))
            RETURN node
            UNION
            MATCH (node:Profile%(node_kind)s)
            WHERE exists((node)-[:HAS_ATTRIBUTE]-(:Attribute { name: $prev_attr.name }))
            RETURN node
            UNION
            MATCH (node:Template%(node_kind)s)
            WHERE exists((node)-[:HAS_ATTRIBUTE]-(:Attribute { name: $prev_attr.name }))
            RETURN node
        }
        WITH node
        """ % {"node_kind": self.previous_attr.node_kind}

        return query

    @staticmethod
    def _render_sub_query_per_rel_type_update_active(rel_type: str, rel_def: FieldInfo) -> str:
        subquery = [
            "WITH peer_node, rb, active_attr",
            f'WHERE type(rb) = "{rel_type}"',
        ]
        if rel_def.default.direction.value == "outbound":
            subquery.append(f"CREATE (active_attr)-[:{rel_type} $rel_props_delete ]->(peer_node)")
        elif rel_def.default.direction.value == "inbound":
            subquery.append(f"CREATE (active_attr)<-[:{rel_type} $rel_props_delete ]-(peer_node)")
        else:
            subquery.append(f"CREATE (active_attr)-[:{rel_type} $rel_props_delete ]-(peer_node)")

        subquery.append("RETURN peer_node as p2")
        return "\n".join(subquery)

    @staticmethod
    def _render_sub_query_per_rel_type_create_new(rel_type: str, rel_def: FieldInfo) -> str:
        subquery = [
            "WITH peer_node, rb, active_attr, new_attr",
            f'WHERE type(rb) = "{rel_type}"',
        ]
        if rel_def.default.direction.value == "outbound":
            subquery.append(f"CREATE (new_attr)-[:{rel_type} $rel_props_create ]->(peer_node)")
        elif rel_def.default.direction.value == "inbound":
            subquery.append(f"CREATE (new_attr)<-[:{rel_type} $rel_props_create ]-(peer_node)")
        else:
            subquery.append(f"CREATE (new_attr)-[:{rel_type} $rel_props_create ]-(peer_node)")

        subquery.append("RETURN peer_node as p2")
        return "\n".join(subquery)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["new_attr"] = self.new_attr.model_dump()
        self.params["prev_attr"] = self.previous_attr.model_dump()

        self.params["current_time"] = self.at.to_string()
        self.params["branch_name"] = self.branch.name

        self.params["user_id"] = self.user_id

        self.params["rel_props_create"] = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
            "from_user_id": self.user_id,
        }

        self.params["rel_props_delete"] = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": RelationshipStatus.DELETED.value,
            "from": self.at.to_string(),
            "from_user_id": self.user_id,
        }

        # Set metadata for vertex properties on default/global branch
        self.params["set_metadata"] = self.branch.is_default or self.branch.is_global

        sub_queries_create = [
            self._render_sub_query_per_rel_type_create_new(rel_type, rel_def)
            for rel_type, rel_def in GraphAttributeRelationships.model_fields.items()
        ]
        sub_query_create_all = "\nUNION\n".join(sub_queries_create)

        sub_queries_update = [
            self._render_sub_query_per_rel_type_update_active(rel_type, rel_def)
            for rel_type, rel_def in GraphAttributeRelationships.model_fields.items()
        ]
        sub_query_update_all = "\nUNION\n".join(sub_queries_update)

        self.add_to_query(self.render_match())

        add_uuid = db.render_uuid_generation(node_label="new_attr", node_attr="uuid")
        query = """
        CALL (node) {
            MATCH (root:Root)<-[r:IS_PART_OF]-(node)
            WHERE %(branch_filter)s
            RETURN node as n1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, r1 as rb
        WHERE rb.status = "active"
        // Find all the attributes that need to be updated
        CALL (active_node) {
            MATCH (active_node)-[r:HAS_ATTRIBUTE]-(attr:Attribute { name: $prev_attr.name })
            WHERE %(branch_filter)s
            RETURN active_node as n1, r as r1, attr as attr1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, r1 as rb, attr1 as active_attr
        WHERE rb.status = "active"
        CREATE (new_attr:Attribute { name: $new_attr.name, branch_support: $new_attr.branch_support })
        %(add_uuid)s
        WITH active_node, active_attr, new_attr
        MATCH (active_attr)-[]-(peer)
        WITH DISTINCT active_node, active_attr, new_attr, peer
        CALL (active_attr, peer) {
            MATCH (active_attr)-[r]-(peer)
            WHERE %(branch_filter)s
            RETURN active_attr as a1, r as r1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH active_node, a1 as active_attr, r1 as rb, p1 as peer_node, new_attr
        WHERE rb.status = "active"
        CALL (peer_node, rb, active_attr, new_attr){
            %(sub_query_create_all)s
        }
        WITH p2 as peer_node, rb, new_attr, active_attr, active_node
        """ % {"branch_filter": branch_filter, "add_uuid": add_uuid, "sub_query_create_all": sub_query_create_all}
        self.add_to_query(query)

        if not (self.branch.is_default or self.branch.is_global):
            query = """
            CALL (peer_node, rb, active_attr) {
                %(sub_query_update_all)s
            }
            WITH p2 as peer_node, rb, new_attr
            RETURN DISTINCT new_attr
            """ % {"sub_query_update_all": sub_query_update_all}
            self.add_to_query(query)
        else:
            query = """
            FOREACH (i in CASE WHEN rb.branch = $branch_name THEN [1] ELSE [] END |
                SET rb.to = $current_time, rb.to_user_id = $user_id
            )
            WITH new_attr, active_node
            // Set metadata on new Attribute and Node vertices if on default/global branch
            CALL (new_attr, active_node) {
                WITH new_attr, active_node
                WHERE $set_metadata
                SET new_attr.created_at = $current_time, new_attr.created_by = $user_id
                SET new_attr.updated_at = $current_time, new_attr.updated_by = $user_id
                SET active_node.updated_at = $current_time, active_node.updated_by = $user_id
            }
            RETURN DISTINCT new_attr
            """
            self.add_to_query(query)



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from infrahub.constants.database import IndexType
from infrahub.core.migrations.shared import MigrationInput, MigrationResult
from infrahub.core.query import Query  # noqa: TC001
from infrahub.database import DatabaseType
from infrahub.database.index import IndexItem
from infrahub.database.neo4j import IndexManagerNeo4j

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


INDEX_TO_DELETE = IndexItem(name="attr_value", label="AttributeValue", properties=["value"], type=IndexType.RANGE)


class Migration014(GraphMigration):
    name: str = "014_remove_index_attr_value"
    queries: Sequence[type[Query]] = []
    minimum_version: int = 13

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        result = MigrationResult()

        # Only execute this migration for Neo4j
        if db.db_type != DatabaseType.NEO4J:
            return result

        try:
            index_manager = IndexManagerNeo4j(db=db)
            index_manager.init(nodes=[INDEX_TO_DELETE], rels=[])
            await index_manager.drop()
        except Exception as exc:
            result.errors.append(str(exc))
            return result

        return result

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()
        return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.types import is_large_attribute_type

from ..query import AttributeMigrationQuery, MigrationBaseQuery
from ..shared import AttributeSchemaMigration, MigrationInput, MigrationResult

if TYPE_CHECKING:
    from infrahub.core.branch.models import Branch
    from infrahub.database import InfrahubDatabase


class AttributeKindUpdateMigrationQuery(AttributeMigrationQuery):
    name = "migration_attribute_kind"
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at)
        self.params.update(branch_params)
        needs_index = not is_large_attribute_type(self.migration.new_attribute_schema.kind)
        self.params["needs_index"] = needs_index
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = self.at.to_string()
        self.params["attr_name"] = self.migration.previous_attribute_schema.name
        self.params["user_id"] = self.user_id

        # Set metadata for vertex properties on default/global branch
        self.params["set_metadata"] = self.branch.is_default or self.branch.is_global

        new_attr_value_labels = "AttributeValue"
        if needs_index:
            new_attr_value_labels += ":AttributeValueIndexed"
        # ruff: noqa: S608
        query = """
// ------------
// start with all the Attribute vertices we might care about
// ------------
MATCH (n:%(schema_kinds)s)-[:HAS_ATTRIBUTE]->(attr:Attribute)
WHERE attr.name = $attr_name
WITH DISTINCT n, attr

// ------------
// for each Attribute, find the most recent active edge and AttributeValue vertex that needs to be [un]indexed
// ------------
CALL (n, attr) {
    MATCH (n)-[r1:HAS_ATTRIBUTE]->(attr:Attribute)-[r2:HAS_VALUE]->(av)
    WHERE all(r IN [r1, r2] WHERE %(branch_filter)s)
    WITH r2, av, r1.status = "active" AND r2.status = "active" AS is_active
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status = "active" DESC, r1.branch_level DESC, r1.from DESC, r1.status = "active" DESC
    LIMIT 1
    WITH r2 AS has_value_e, av, "AttributeValueIndexed" IN labels(av) AS is_indexed
    WHERE is_active AND is_indexed <> $needs_index
    RETURN has_value_e, av
}

// ------------
// check if the correct AttributeValue vertex to use exists
// create it if not
// ------------
WITH DISTINCT av.is_default AS av_is_default, av.value AS av_value
CALL (av_is_default, av_value) {
    OPTIONAL MATCH (existing_av:AttributeValue {is_default: av_is_default, value: av_value})
    WHERE "AttributeValueIndexed" IN labels(existing_av) = $needs_index
    WITH existing_av WHERE existing_av IS NULL
    LIMIT 1
    CREATE (:%(new_attr_value_labels)s {is_default: av_is_default, value: av_value})
}

// ------------
// get all the AttributeValue vertices that need to be updated again and run the updates
// ------------
WITH 1 AS one
LIMIT 1
MATCH (n:%(schema_kinds)s)-[:HAS_ATTRIBUTE]->(attr:Attribute)
WHERE attr.name = $attr_name
WITH DISTINCT n, attr

// ------------
// for each Attribute, find the most recent active edge and AttributeValue vertex that needs to be [un]indexed
// ------------
CALL (n, attr) {
    MATCH (n)-[r1:HAS_ATTRIBUTE]->(attr:Attribute)-[r2:HAS_VALUE]->(av)
    WHERE all(r IN [r1, r2] WHERE %(branch_filter)s)
    WITH r2, av, r1.status = "active" AND r2.status = "active" AS is_active
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status = "active" DESC, r1.branch_level DESC, r1.from DESC, r1.status = "active" DESC
    LIMIT 1
    WITH r2 AS has_value_e, av, "AttributeValueIndexed" IN labels(av) AS is_indexed
    WHERE is_active AND is_indexed <> $needs_index
    RETURN has_value_e, av
}

// ------------
// create and update the HAS_VALUE edges
// ------------
CALL (attr, has_value_e, av) {
    // ------------
    // get the correct AttributeValue vertex b/c it definitely exists now
    // ------------
    MATCH (new_av:%(new_attr_value_labels)s {is_default: av.is_default, value: av.value})
    WHERE "AttributeValueIndexed" IN labels(new_av) = $needs_index
    LIMIT 1

    // ------------
    // create the new HAS_VALUE edge
    // ------------
    CREATE (attr)-[new_has_value_e:HAS_VALUE]->(new_av)
    SET new_has_value_e = properties(has_value_e)
    SET new_has_value_e.status = "active"
    SET new_has_value_e.branch = $branch
    SET new_has_value_e.branch_level = $branch_level
    SET new_has_value_e.from = $at
    SET new_has_value_e.from_user_id = $user_id
    SET new_has_value_e.to = NULL

    // ------------
    // if we are updating on a branch and the existing edge is on the default branch,
    // then create a new deleted edge on this branch
    // ------------
    WITH attr, has_value_e, av
    WHERE has_value_e.branch <> $branch
    CREATE (attr)-[deleted_has_value_e:HAS_VALUE]->(av)
    SET deleted_has_value_e = properties(has_value_e)
    SET deleted_has_value_e.status = "deleted"
    SET deleted_has_value_e.branch = $branch
    SET deleted_has_value_e.branch_level = $branch_level
    SET deleted_has_value_e.from = $at
    SET deleted_has_value_e.from_user_id = $user_id
    SET deleted_has_value_e.to = NULL
}

// ------------
// if the existing edge is on the same branch as the update,
// then set its "to" time
// ------------
CALL (has_value_e) {
    WITH has_value_e
    WHERE has_value_e.branch = $branch
    SET has_value_e.to = $at, has_value_e.to_user_id = $user_id
}

// ------------
// Set metadata on Attribute and Node vertices if on default/global branch
// ------------
CALL (attr, n) {
    WITH attr, n
    WHERE $set_metadata
    SET attr.updated_at = $at, attr.updated_by = $user_id
    SET n.updated_at = $at, n.updated_by = $user_id
}
        """ % {
            "schema_kinds": (
                f"{self.migration.previous_schema.kind}|Profile{self.migration.previous_schema.kind}|Template{self.migration.previous_schema.kind}"
            ),
            "branch_filter": branch_filter,
            "new_attr_value_labels": new_attr_value_labels,
        }
        self.add_to_query(query)


class AttributeKindUpdateMigration(AttributeSchemaMigration):
    name: str = "attribute.kind.update"
    queries: Sequence[type[AttributeMigrationQuery]] = [AttributeKindUpdateMigrationQuery]  # type: ignore[assignment]

    async def execute(
        self,
        migration_input: MigrationInput,
        branch: Branch,
        queries: Sequence[type[MigrationBaseQuery]] | None = None,
    ) -> MigrationResult:
        is_indexed_previous = is_large_attribute_type(self.previous_attribute_schema.kind)
        is_indexed_new = is_large_attribute_type(self.new_attribute_schema.kind)
        if is_indexed_previous is is_indexed_new:
            return MigrationResult()

        return await super().execute(migration_input=migration_input, branch=branch, queries=queries)



from __future__ import annotations

from typing import Self

from pydantic import BaseModel, model_validator

from infrahub.constants.enums import OrderDirection  # noqa: TC001
from infrahub.exceptions import ValidationError

# Metadata field name constants
METADATA_CREATED_AT = "created_at"
METADATA_CREATED_BY = "created_by"
METADATA_UPDATED_AT = "updated_at"
METADATA_UPDATED_BY = "updated_by"


class NodeMetaOrder(BaseModel):
    created_at: OrderDirection | None = None
    updated_at: OrderDirection | None = None


class OrderModel(BaseModel):
    disable: bool | None = None
    node_metadata: NodeMetaOrder | None = None

    @model_validator(mode="after")
    def validate_metadata(self) -> Self:
        if self.node_metadata and self.node_metadata.created_at and self.node_metadata.updated_at:
            raise ValidationError("Cannot order by both created_at and updated_at simultaneously.")
        return self



from __future__ import annotations

from typing import TYPE_CHECKING

import ujson
from infrahub_sdk import InfrahubClient  # noqa: TC002  needed for prefect flow
from infrahub_sdk.protocols import CoreTransformPython, CoreWebhook
from prefect import flow, task
from prefect.automations import AutomationCore
from prefect.cache_policies import NONE
from prefect.client.orchestration import get_client as get_prefect_client
from prefect.logging import get_run_logger

from infrahub.message_bus.types import KVTTL
from infrahub.trigger.models import ExecuteWorkflow, TriggerType
from infrahub.trigger.setup import gather_all_automations, setup_triggers_specific
from infrahub.workers.dependencies import get_cache, get_client, get_database, get_http
from infrahub.workflows.utils import add_tags

from .gather import gather_trigger_webhook
from .models import CustomWebhook, EventContext, StandardWebhook, TransformWebhook, Webhook, WebhookTriggerDefinition

if TYPE_CHECKING:
    from httpx import Response


WEBHOOK_MAP: dict[str, type[Webhook]] = {
    "StandardWebhook": StandardWebhook,
    "CustomWebhook": CustomWebhook,
    "TransformWebhook": TransformWebhook,
}


@task(name="webhook-send", task_run_name="Send Standard Webhook {webhook.name}", cache_policy=NONE, retries=3)
async def webhook_send(webhook: Webhook, context: EventContext, event_data: dict) -> Response:
    http_service = get_http()
    client = get_client()
    response = await webhook.send(data=event_data, context=context, http_service=http_service, client=client)
    response.raise_for_status()
    return response


@task(name="webhook-convert-node", task_run_name="Convert node to webhook", cache_policy=NONE)
async def convert_node_to_webhook(webhook_node: CoreWebhook, client: InfrahubClient) -> Webhook:
    webhook_kind = webhook_node.get_kind()

    if webhook_kind not in ["CoreStandardWebhook", "CoreCustomWebhook"]:
        raise ValueError(f"Unsupported webhook kind: {webhook_kind}")

    if webhook_kind == "CoreStandardWebhook":
        return StandardWebhook.from_object(obj=webhook_node)

    # Processing Custom Webhook
    if webhook_node.transformation.id:
        transform = await client.get(
            kind=CoreTransformPython,
            id=webhook_node.transformation.id,
            prefetch_relationships=True,
            include=["name", "class_name", "file_path", "repository"],
        )
        return TransformWebhook.from_object(obj=webhook_node, transform=transform)

    return CustomWebhook.from_object(obj=webhook_node)


@flow(name="webhook-process", flow_run_name="Send webhook for {webhook_name}")
async def webhook_process(
    webhook_id: str,
    webhook_name: str,  # noqa: ARG001
    webhook_kind: str,
    event_id: str,
    event_type: str,
    event_occured_at: str,
    event_payload: dict,
    branch_name: str | None = None,
) -> None:
    log = get_run_logger()
    client = get_client()
    cache = await get_cache()

    if branch_name:
        await add_tags(branches=[branch_name])

    webhook_data_str = await cache.get(key=f"webhook:{webhook_id}")
    if not webhook_data_str:
        log.info(f"Webhook {webhook_id} not found in cache")
        webhook_node = await client.get(kind=webhook_kind, id=webhook_id)
        webhook = await convert_node_to_webhook(webhook_node=webhook_node, client=client)
        webhook_data = webhook.to_cache()
        await cache.set(key=f"webhook:{webhook_id}", value=ujson.dumps(webhook_data), expires=KVTTL.TWO_HOURS)

    else:
        webhook_data = ujson.loads(webhook_data_str)

        if webhook_data["webhook_type"] not in WEBHOOK_MAP:
            raise ValueError(f"Unsupported webhook kind: {webhook_data['webhook_type']}")

        webhook_class = WEBHOOK_MAP[webhook_data["webhook_type"]]
        webhook = webhook_class.from_cache(webhook_data)

    webhook_context = EventContext.from_event(
        event_id=event_id,
        event_type=event_type,
        event_occured_at=event_occured_at,
        event_payload=event_payload,
    )
    event_data = event_payload.get("data", {})
    response = await webhook_send(webhook=webhook, context=webhook_context, event_data=event_data)
    log.info(f"Successfully sent webhook to {response.url} with status {response.status_code}")


@flow(name="webhook-setup-automation-all", flow_run_name="Configure all webhooks")
async def configure_webhook_all() -> None:
    log = get_run_logger()

    database = await get_database()
    async with database.start_session(read_only=True) as db:
        triggers = await gather_trigger_webhook(db=db)

    log.info(f"{len(triggers)} Webhooks automation configuration completed")
    await setup_triggers_specific(gatherer=gather_trigger_webhook, db=database, trigger_type=TriggerType.WEBHOOK)  # type: ignore[arg-type]


@flow(name="webhook-setup-automation-one", flow_run_name="Configurate webhook for {webhook_name}")
async def configure_webhook_one(
    webhook_name: str,
    event_data: dict,
) -> None:
    log = get_run_logger()

    webhook = await get_client().get(kind=CoreWebhook, id=event_data["node_id"])
    trigger = WebhookTriggerDefinition.from_object(webhook)

    async with get_prefect_client(sync_client=False) as prefect_client:
        all_automations = await gather_all_automations(client=prefect_client)
        existing_automations = [
            automation for automation in all_automations if automation.name == trigger.generate_name()
        ]
        existing_automation = existing_automations[0] if existing_automations else None

        # If webhook is inactive, delete the automation if it exists
        if not webhook.active.value:
            if existing_automation:
                await prefect_client.delete_automation(automation_id=existing_automation.id)
                log.info(f"Automation {trigger.generate_name()} deleted (webhook disabled)")
            else:
                log.info(f"Webhook {webhook_name} is disabled, no automation to delete")

            cache = await get_cache()
            await cache.delete(key=f"webhook:{webhook.id}")
            return

        # Query the deployment associated with the trigger to have its ID
        deployment_name = trigger.get_deployment_names()[0]
        deployment = await prefect_client.read_deployment_by_name(name=f"{deployment_name}/{deployment_name}")

        automation = AutomationCore(
            name=trigger.generate_name(),
            description=trigger.get_description(),
            enabled=True,
            trigger=trigger.trigger.get_prefect(),
            actions=[action.get(deployment.id) for action in trigger.actions if isinstance(action, ExecuteWorkflow)],
        )

        if existing_automation:
            await prefect_client.update_automation(automation_id=existing_automation.id, automation=automation)
            log.info(f"Automation {trigger.generate_name()} updated")
        else:
            await prefect_client.create_automation(automation=automation)
            log.info(f"Automation {trigger.generate_name()} created")

        cache = await get_cache()
        await cache.delete(key=f"webhook:{webhook.id}")


@flow(name="webhook-delete-automation", flow_run_name="Delete webhook automation for {webhook_name}")
async def delete_webhook_automation(
    webhook_id: str,
    webhook_name: str,  # noqa: ARG001
) -> None:
    log = get_run_logger()

    async with get_prefect_client(sync_client=False) as prefect_client:
        automation_name = WebhookTriggerDefinition.generate_name_from_id(id=webhook_id)

        all_automations = await gather_all_automations(client=prefect_client)
        existing_automations = [automation for automation in all_automations if automation.name == automation_name]
        existing_automation = existing_automations[0] if existing_automations else None

        if existing_automation:
            await prefect_client.delete_automation(automation_id=existing_automation.id)
            log.info(f"Automation {automation_name} deleted")

        cache = await get_cache()
        await cache.delete(key=f"webhook:{webhook_id}")
